import sys
from typing import Annotated

import uvicorn
from fastapi import Body, FastAPI, Request
from fastapi.responses import JSONResponse

from .database import json_decode, json_encode, populse_db_table
from .storage_api import StorageFileAPI, serialize_exception

body_str = Annotated[str, Body(embed=True)]
body_path = Annotated[list[str | int | list[str]], Body(embed=True)]
body_bool = Annotated[bool, Body()]
body_dict = Annotated[dict, Body()]
body_json = Annotated[str | int | float | bool | None | list | dict, Body()]


def create_server(database_file, create=True):
    storage_api = StorageFileAPI(database_file, create=create)
    app = FastAPI()

    @app.middleware("http")
    async def cors_middleware(request: Request, call_next):
        
        response = await call_next(request)
        response.headers["Access-Control-Allow-Origin"] = "*"
        response.headers["Access-Control-Allow-Methods"] = "*"
        return response

    @app.middleware("http")
    async def catch_exceptions_middleware(request: Request, call_next):
        try:
            return await call_next(request)
        except Exception as exc:
            return JSONResponse(
                status_code=500,
                content=serialize_exception(exc),
            )

    @app.get("/access_token")
    async def access_token():
        # TODO: give a real re/write challenge to the user in
        # order to get its access rights.
        return storage_api.access_token()

    @app.post("/connection")
    async def connect(
        access_token: body_str,
        exclusive: body_bool,
        write: body_bool,
        create: body_bool,
    ):
        return storage_api.connect(access_token, exclusive, write, create)

    @app.delete("/connection")
    async def disconnect(connection_id: body_str, rollback: body_bool):
        return storage_api.disconnect(connection_id, rollback)

    @app.post("/schema_collection")
    async def add_schema_collections(
        connection_id: body_str, schema_to_collections: body_dict
    ):
        return storage_api.add_schema_collections(connection_id, schema_to_collections)

    @app.post("/schema/{name}")
    async def add_collection(
        connection_id: body_str,
        name: str,
        primary_key: Annotated[str | list[str] | dict[str, str], Body()],
    ):
        return storage_api.add_collection(connection_id, name, primary_key)

    @app.post("/schema/{collection_name}/{field_name}")
    async def add_field(
        connection_id: body_str,
        collection_name: str,
        field_name: str,
        field_type: body_str,
        description: Annotated[str | None, Body()] = None,
        index: body_bool = False,
    ):
        return storage_api.add_field(
            connection_id,
            collection_name,
            field_name,
            field_type,
            description,
            index,
        )

    @app.delete("/schema/{collection_name}/{field_name}")
    async def remove_field(
        connection_id: body_str,
        collection_name: str,
        field_name: str,
    ):
        return storage_api.remove_field(
            connection_id,
            collection_name,
            field_name,
        )

    @app.get("/data")
    async def get(
        connection_id: body_str,
        path: body_path,
        default: body_json = None,
        fields: Annotated[list[str] | None, Body()] = None,
        as_list: body_bool = False,
        distinct: body_bool = False,
    ):
        result = storage_api.get(
            connection_id,
            path,
            default,
            fields,
            as_list,
            distinct,
        )
        return json_encode(result)

    @app.get("/count")
    async def count(
        connection_id: body_str,
        path: body_path,
        query: Annotated[str | None, Body()] = None,
    ):
        return storage_api.count(connection_id, path, query)

    @app.get("/primary_key")
    async def primary_key(connection_id: body_str, path: body_path):
        return storage_api.primary_key(connection_id, path)

    @app.post("/data")
    async def set(connection_id: body_str, path: body_path, value: body_json):
        return storage_api.set(connection_id, path, json_decode(value))

    @app.delete("/data")
    async def delete(connection_id: body_str, path: body_path):
        return storage_api.delete(connection_id, path)

    @app.put("/data")
    async def update(connection_id: body_str, path: body_path, value: body_json):
        return storage_api.update(connection_id, path, json_decode(value))

    @app.patch("/data")
    async def append(connection_id: body_str, path: body_path, value: body_json):
        return storage_api.append(connection_id, path, json_decode(value))

    @app.get("/search")
    async def search(
        connection_id: body_str,
        path: body_path,
        query: Annotated[str | None, Body()] = None,
        fields: Annotated[list[str] | None, Body()] = None,
        as_list: body_bool = False,
        distinct: body_bool = False,
    ):
        result = storage_api.search(
            connection_id, path, query, fields, as_list, distinct
        )
        return json_encode(result)

    @app.delete("/search")
    async def search_and_delete(
        connection_id: body_str,
        path: body_path,
        query: Annotated[str | None, Body()] = None,
    ):
        return storage_api.search_and_delete(connection_id, path, query)

    @app.get("/distinct")
    async def distinct_values(
        connection_id: body_str,
        path: body_path,
        field: body_str,
    ):
        result = storage_api.distinct_values(connection_id, path, field)
        return json_encode(result)

    @app.delete("/")
    async def clear_database(
        connection_id: body_str,
        path: body_path,
    ):
        return storage_api.clear_database(connection_id, path)

    @app.get("/has_collection")
    async def has_collection(
        connection_id: body_str,
        path: body_path,
        collection: body_str,
    ):
        return storage_api.has_collection(connection_id, path, collection)

    @app.get("/collection_names")
    async def collection_names(
        connection_id: body_str,
        path: body_path,
    ):
        return storage_api.collection_names(connection_id, path)

    @app.get("/keys")
    async def keys(
        connection_id: body_str,
        path: body_path,
    ):
        result = storage_api.keys(connection_id, path)
        return list(result)

    return app


if __name__ == "__main__":
    import sqlite3

    database_file = sys.argv[1]
    if len(sys.argv) > 2:
        port = int(sys.argv[2])
    else:
        port = 5000
    cnx = sqlite3.connect(database_file, isolation_level="EXCLUSIVE", timeout=10)
    try:
        sql = (
            "CREATE TABLE IF NOT EXISTS "
            f"[{populse_db_table}] ("
            "category TEXT NOT NULL,"
            "key TEXT NOT NULL,"
            "_json TEXT,"
            "PRIMARY KEY (category, key))"
        )
        cnx.execute(sql)
        row = cnx.execute(
            f"SELECT _json FROM [{populse_db_table}] WHERE category='server' AND key='url'"
        ).fetchone()
        if row:
            raise RuntimeError(f"{database_file} already have a server in {row[0]}")
        url = f"http://0.0.0.0:{port}"
        cnx.execute(
            f"INSERT INTO [{populse_db_table}] (category, key, _json) VALUES (?,?,?)",
            ["server", "url", url],
        )
        cnx.commit()
    finally:
        cnx.close()
    try:
        app = create_server(database_file)
        uvicorn.run(app, port=port, log_level="critical")
    finally:
        cnx = sqlite3.connect(database_file, isolation_level="EXCLUSIVE", timeout=10)
        try:
            cnx.execute(
                f"DELETE FROM [{populse_db_table}] WHERE category='server' AND key='url'"
            )
            cnx.commit()
        finally:
            cnx.close()



import functools
import json
import os
import sqlite3
from datetime import date, datetime, time

import dateutil

from ..database import (
    DatabaseCollection,
    DatabaseSession,
    json_decode,
    json_dumps,
    json_encode,
    populse_db_table,
    str_to_type,
    type_to_sqlite,
    type_to_str,
)
from ..filter import FilterToSQL, filter_parser

"""
SQLite3 implementation of populse_db engine.

A populse_db engine is created when a DatabaseSession object is created
(typically within a "with" statement)
"""

if tuple(int(i) for i in sqlite3.sqlite_version.split(".")) < (3, 38, 0):
    raise NotImplementedError(
        f"populse_db requires a SQLite version > 3.38.0 but current version is {sqlite3.sqlite_version}"
    )
sqlite3.register_adapter(datetime, lambda d: d.isoformat())
sqlite3.register_adapter(date, lambda d: d.isoformat())
sqlite3.register_adapter(time, lambda d: d.isoformat())
sqlite3.register_converter("bool", lambda b: bool(int(b)))
sqlite3.register_converter("datetime", lambda b: dateutil.parser.parse(b))
sqlite3.register_converter("date", lambda b: dateutil.parser.parse(b).date())
sqlite3.register_converter("time", lambda b: dateutil.parser.parse(b).time())


def create_sqlite_session_factory(url):
    if url.path:
        sqlite_file = url.path
    elif url.netloc:
        sqlite_file = url.netloc
    result = functools.partial(sqlite_session_factory, sqlite_file)
    return result


def sqlite_session_factory(sqlite_file, *args, create=False, **kwargs):
    if not create and not os.path.exists(sqlite_file):
        return None
    return SQLiteSession(sqlite_file, *args, **kwargs)


class ParsedFilter(str):
    pass


class SQLiteSession(DatabaseSession):
    database_exceptions = (
        sqlite3.OperationalError,
        sqlite3.IntegrityError,
    )

    def __init__(self, sqlite_file, exclusive=False, timeout=None, echo_sql=None):
        self.echo_sql = echo_sql
        self.sqlite = sqlite3.connect(
            sqlite_file,
            isolation_level=None,
            check_same_thread=False,
            detect_types=sqlite3.PARSE_DECLTYPES,
        )
        self.exclusive = exclusive
        if timeout:
            self.sqlite.execute(f"PRAGMA busy_timeout={timeout}")
        self.sqlite.executescript(
            "PRAGMA synchronous=OFF;"
            "PRAGMA case_sensitive_like=ON;"
            "PRAGMA foreign_keys=ON;"
            f"BEGIN {('EXCLUSIVE' if self.exclusive else 'DEFERRED')};"
        )
        self._collection_cache = {}
        # Iterate on all collections to put them in cache
        all(self)

    def close(self, rollback=False):
        if rollback:
            self.sqlite.rollback()
        else:
            self.sqlite.commit()
        self.sqlite.close()

    def has_collection(self, name):
        return name in self._collection_cache

    def get_collection(self, name):
        return self._collection_cache.get(name)

    def __getitem__(self, collection_name):
        result = self._collection_cache.get(collection_name)
        if result is None:
            result = SQLiteCollection(self, collection_name)
            self._collection_cache[collection_name] = result
        return result

    def execute(self, sql, data=None):
        try:
            if data:
                result = self.sqlite.execute(sql, data)
                if self.echo_sql:
                    print(sql, data, file=self.echo_sql)
                return result
            else:
                result = self.sqlite.execute(sql)
                if self.echo_sql:
                    print(sql, file=self.echo_sql)
                return result
        except sqlite3.OperationalError as e:
            raise sqlite3.OperationalError(f"Error in SQL request: {sql}") from e

    def commit(self):
        self.sqlite.commit()
        self.sqlite.execute(f"BEGIN {('EXCLUSIVE' if self.exclusive else 'DEFERRED')}")

    def rollback(self):
        self.sqlite.rollback()
        self.sqlite.execute(f"BEGIN {('EXCLUSIVE' if self.exclusive else 'DEFERRED')}")

    def settings(self, category, key, default=None):
        try:
            sql = f"SELECT _json FROM [{populse_db_table}] WHERE category=? and key=?"
            cur = self.execute(sql, [category, key])
        except sqlite3.OperationalError:
            return default
        j = cur.fetchone()
        if j:
            return json.loads(j[0])
        return default

    def set_settings(self, category, key, value):
        sql = f"INSERT OR REPLACE INTO {populse_db_table} (category, key, _json) VALUES (?,?,?)"
        data = [category, key, json_dumps(value)]
        retry = False
        try:
            self.execute(sql, data)
        except sqlite3.OperationalError:
            retry = True
        if retry:
            sql2 = (
                "CREATE TABLE IF NOT EXISTS "
                f"[{populse_db_table}] ("
                "category TEXT NOT NULL,"
                "key TEXT NOT NULL,"
                "_json TEXT,"
                "PRIMARY KEY (category, key))"
            )
            self.execute(sql2)
            self.execute(sql, data)

    def clear(self):
        """
        Erase the whole database content.
        """
        sql = 'SELECT name FROM sqlite_master WHERE type = "table"'
        tables = [i[0] for i in self.execute(sql)]
        for table in tables:
            sql = f"DROP TABLE [{table}]"
            self.execute(sql)
        self._collection_cache = {}

    def add_collection(
        self,
        name,
        primary_key=DatabaseSession.default_primary_key,
        catchall_column="_catchall",
    ):
        if isinstance(primary_key, str):
            dict_primary_key = {primary_key: "str"}
        elif isinstance(primary_key, list | tuple):
            dict_primary_key = {i: "str" for i in primary_key}
        else:
            dict_primary_key = {
                k: (
                    type_to_sqlite(str_to_type(v))
                    if isinstance(v, str)
                    else type_to_str(v)
                )
                for k, v in primary_key.items()
            }
        sql = (
            f"CREATE TABLE [{name}] ("
            f"{','.join(f'[{n}] {t} NOT NULL' for n, t in dict_primary_key.items())},"
            f"{catchall_column} dict,"
            f"PRIMARY KEY ({','.join(f'[{i}]' for i in dict_primary_key.keys())}))"
        )
        self.execute(sql)
        # Accessing the collection to put it in cache
        self[name]

    def remove_collection(self, name):
        sql = f"DROP TABLE [{name}]"
        self.execute(sql)
        self._collection_cache.pop(name, None)

    def __iter__(self):
        sql = "SELECT name FROM sqlite_master WHERE type='table'"
        for row in self.execute(sql):
            table = row[0]
            if table == populse_db_table:
                continue
            yield self[table]


class SQLiteCollection(DatabaseCollection):
    _column_encodings = {
        list: (
            lambda l: (None if l is None else json_dumps(l)),  # noqa: E741
            lambda l: (None if l is None else json.loads(l)),  # noqa: E741
        ),
        dict: (
            lambda d: (None if d is None else json_dumps(d)),
            lambda d: (None if d is None else json.loads(d)),
        ),
    }

    def __init__(self, session, name):
        super().__init__(session, name)
        settings = self.session.settings("collection", name, {})
        sql = f"pragma table_info([{self.name}])"
        bad_table = True
        catchall_column_found = False
        for row in self.session.execute(sql):
            bad_table = False
            if row[1] == self.catchall_column:
                catchall_column_found = True
                continue
            column_type_str = row[2].lower()
            column_type = str_to_type(column_type_str)
            main_type = getattr(column_type, "__origin__", None) or column_type
            encoding = self._column_encodings.get(main_type)
            if row[5]:
                self.primary_key[row[1]] = column_type
            field = {
                "collection": self.name,
                "name": row[1],
                "primary_key": bool(row[5]),
                "type": column_type,
                "encoding": encoding,
            }
            field_settings = settings.get("fields", {}).get(row[1], {})
            field.update(field_settings)
            if field_settings.get("bad_json", False):
                self.bad_json_fields.add(row[1])
            self.fields[row[1]] = field
        if bad_table:
            raise ValueError(f"No such database table: {name}")
        if self.catchall_column and not catchall_column_found:
            raise ValueError(f"table {name} must have a column {self.catchall_column}")

    def add_field(
        self, name, field_type, description=None, index=False, bad_json=False
    ):
        if isinstance(field_type, str):
            field_type = str_to_type(field_type)
        sql = f"ALTER TABLE [{self.name}] ADD COLUMN [{name}] {type_to_sqlite(field_type)}"
        self.session.execute(sql)
        if index:
            sql = f"CREATE INDEX [{self.name}_{name}] ON [{self.name}] ([{name}])"
            self.session.execute(sql)
        settings = self.settings()
        settings.setdefault("fields", {})[name] = {
            "description": description,
            "index": index,
            "bad_json": bad_json,
        }
        self.set_settings(settings)
        field = {
            "collection": self.name,
            "name": name,
            "primary_key": False,
            "type": field_type,
            "description": description,
            "index": index,
            "bad_json": bad_json,
            "encoding": self._column_encodings.get(
                getattr(field_type, "__origin__", None) or field_type
            ),
        }
        self.fields[name] = field
        if bad_json:
            self.bad_json_fields.add(name)

    def remove_field(self, name):
        """
        Removes a specified field from the table and updates associated
        metadata.

        Args:
            name (str): The name of the field to remove.

        Raises:
            ValueError: If attempting to remove a primary key field.
            NotImplementedError: If the SQLite version is below 3.35.0,
                                 which does not support removing columns.
        """
        if name in self.primary_key:
            raise ValueError("Cannot remove a key field")

        sql = f"ALTER TABLE [{self.name}] DROP COLUMN [{name}]"
        self.session.execute(sql)
        # Update metadata
        settings = self.settings()
        fields = settings.setdefault("fields", {})
        fields.pop(name, None)
        self.set_settings(settings)
        self.fields.pop(name, None)
        self.bad_json_fields.discard(name)

    def has_document(self, document_id):
        document_id = self.document_id(document_id)
        sql = f"SELECT count(*) FROM [{self.name}] WHERE {' AND '.join(f'[{i}] = ?' for i in self.primary_key)}"
        return next(self.session.execute(sql, document_id))[0] != 0

    def _documents(self, where, where_data, fields, as_list, distinct):
        json_decode_columns = []
        if fields:
            columns = []
            catchall_fields = False
            for field in fields:
                if field in self.fields:
                    columns.append(f"[{field}]")
                else:
                    json_decode_columns.append(len(columns))
                    columns.append(f"[{self.catchall_column}] -> '$.{field}'")
        else:
            fields = self.fields
            columns = [f"[{i}]" for i in fields]
            catchall_fields = bool(self.catchall_column)
            if catchall_fields:
                columns.append(f"[{self.catchall_column}] -> '$'")
                if as_list:
                    raise ValueError(
                        f"as_list=True cannot be used on {self.name} without a fields list because two documents can have different fields"
                    )

        sql = f"SELECT {('DISTINCT ' if distinct else '')}{','.join(columns)} FROM [{self.name}]"
        if where:
            sql += f" WHERE {where}"
        cur = self.session.execute(sql, where_data)
        for row in cur:
            for i in json_decode_columns:
                if row[i] is None:
                    continue
                if isinstance(row, tuple):
                    row = list(row)
                row[i] = json.loads(row[i])
            if catchall_fields:
                if row[-1] is not None:
                    catchall = json.loads(row[-1])
                    if isinstance(catchall_fields, set):
                        catchall = {i: catchall[i] for i in catchall_fields}
                else:
                    catchall = {}
                row = row[:-1]
            else:
                catchall = {}
            if columns[-1] == self.catchall_column:
                columns = columns[:-1]
            document = json_decode(catchall)
            if isinstance(document, dict):
                document.update(zip(fields, row, strict=True))
                for field, value in document.items():
                    encoding = self.fields.get(field, {}).get("encoding")
                    if encoding:
                        encode, decode = encoding
                        value = decode(value)
                    if field in self.bad_json_fields:
                        value = json_decode(value)
                    document[field] = value
                if as_list:
                    yield [document[i] for i in fields]
                else:
                    yield document
            else:
                yield document

    def count(self, filter=None):
        where = self.parse_filter(filter)
        sql = f"SELECT COUNT(*) FROM [{self.name}]"
        if where:
            sql += f" WHERE {where}"
        return self.session.execute(sql, None).fetchone()[0]

    def document(self, document_id, fields=None, as_list=False):
        document_id = self.document_id(document_id)
        where = f"{' AND '.join(f'[{i}] = ?' for i in self.primary_key)}"
        try:
            return next(self._documents(where, document_id, fields, as_list, False))
        except StopIteration:
            return None

    def documents(self, fields=None, as_list=False, distinct=False):
        yield from self._documents(None, None, fields, as_list, distinct)

    def add(self, document, replace=False):
        document_id = tuple(document.get(i) for i in self.primary_key)
        self._set_document(document_id, document, replace=replace)

    def __setitem__(self, document_id, document):
        document_id = self.document_id(document_id)
        self._set_document(document_id, document, replace=True)

    def _dict_to_sql_update(self, document):
        columns = []
        data = []
        catchall_column = None
        catchall_data = None
        catchall = ...
        if isinstance(document, dict):
            catchall = {}
            for field, value in document.items():
                if field in self.primary_key:
                    continue
                if field in self.bad_json_fields:
                    value = json_encode(value)
                if field in self.fields:
                    columns.append(field)
                    data.append(self._encode_column_value(field, value))
                else:
                    catchall[field] = value
        else:
            catchall_column = self.catchall_column
            catchall_data = document
        if catchall is not ...:
            if not self.catchall_column:
                raise ValueError(
                    f"Collection {self.name} cannot store this value: {catchall}"
                )
            bad_json = False
            try:
                catchall_data = catchall
                json_dumps(catchall)
            except TypeError:
                if isinstance(catchall, dict):
                    bad_json = True
                else:
                    raise
            if bad_json:
                jsons = {}
                for field, value in catchall.items():
                    bad_json = False
                    try:
                        j = value
                        json_dumps(value)
                    except TypeError:
                        bad_json = True
                    if bad_json:
                        self.add_field(field, dict, bad_json=True)
                        column_value = self._encode_column_value(field, value)
                        columns.append(field)
                        data.append(column_value)
                    else:
                        jsons[field] = j
                if jsons:
                    catchall_column = self.catchall_column
                    catchall_data = jsons

            else:
                catchall_column = self.catchall_column
        return columns, data, catchall_column, catchall_data

    def _set_document(self, document_id, document, replace):
        columns, data, catchall_column, catchall_data = self._dict_to_sql_update(
            document
        )
        catchall_data = json.dumps(json_encode(catchall_data))
        columns = [i for i in self.primary_key] + columns
        data = [i for i in document_id] + data
        if catchall_column:
            columns.append(catchall_column)
            data.append(catchall_data)
        if replace:
            replace = " OR REPLACE"
        else:
            replace = ""
        sql = f"INSERT{replace} INTO [{self.name}] ({','.join(f'[{i}]' for i in columns)}) values ({','.join('?' for i in data)})"
        self.session.execute(sql, data)

    def update_document(self, document_id, partial_document):
        document_id = self.document_id(document_id)
        if not all(
            y is None or x == y
            for x, y in zip(
                document_id,
                (partial_document.get(i) for i in self.primary_key),
                strict=True,
            )
        ):
            raise ValueError("Modification of a document's primary key is not allowed")
        columns, data, catchall_column, catchall_data = self._dict_to_sql_update(
            partial_document
        )

        if catchall_column and catchall_data:
            catchall_update = [
                f'[{catchall_column}]=json_set(IFNULL([{catchall_column}],"{{}}"),{",".join("?,json(?)" for i in catchall_data)})'
            ]
            for k, v in catchall_data.items():
                data.append(f"$.{k}")
                data.append(json.dumps(v))
        else:
            catchall_update = []
        where = " AND ".join(f"[{i}]=?" for i in self.primary_key)
        data = data + [i for i in document_id]
        affectations = [f"[{i}]=?" for i in columns] + catchall_update
        if not affectations:
            return
        sql = f"UPDATE [{self.name}] SET {','.join(affectations)} WHERE {where}"
        cur = self.session.execute(sql, data)
        if not cur.rowcount:
            raise ValueError(f"Document with key {document_id} does not exist")

    def __delitem__(self, document_id):
        document_id = self.document_id(document_id)
        sql = f"DELETE FROM [{self.name}] WHERE {' AND '.join(f'[{i}] = ?' for i in self.primary_key)}"
        self.session.execute(sql, document_id)

    def parse_filter(self, filter):
        if filter is None or isinstance(filter, ParsedFilter):
            return filter
        tree = filter_parser().parse(filter)
        where_filter = FilterToSQL(self).transform(tree)
        if where_filter is None:
            return None
        else:
            return ParsedFilter(" ".join(where_filter))

    def filter(self, filter, fields=None, as_list=False, distinct=False):
        parsed_filter = self.parse_filter(filter)
        yield from self._documents(
            parsed_filter, None, fields=fields, as_list=as_list, distinct=distinct
        )

    def delete(self, filter):
        where = self.parse_filter(filter)
        sql = f"DELETE FROM [{self.name}]"
        if where:
            sql += f" WHERE {where}"
        cur = self.session.execute(sql)
        return cur.rowcount



import importlib
import pathlib
import types
import typing
from contextlib import contextmanager

from .database import type_to_str
from .storage_api import StorageAPI


class Storage:
    default_collection = "_"
    schema_collection = "_schema"
    default_field = "_"
    default_document_id = "_"

    def __init__(
        self,
        database_file: str | pathlib.Path,
        timeout: float | None = 10000,
        create: bool = False,
        echo_sql: typing.TextIO | None = None,
    ):
        if isinstance(database_file, pathlib.Path):
            database_file = str(database_file)
        self.storage_api = StorageAPI(
            database_file, timeout=timeout, create=create, echo_sql=echo_sql
        )
        self.access_token = self.storage_api.access_token()
        self._current_data_session = None

    @contextmanager
    def data(self, exclusive=None, write=False, create=False):
        if self._current_data_session is not None:
            storage_session, is_exclusive, is_write = self._current_data_session
            if exclusive and is_exclusive is not True:
                raise RuntimeError(
                    "Impossible to get an exclusive data session because another non exclusive data session exists"
                )
            if write and not is_write:
                raise RuntimeError(
                    "Impossible to get an write data session because another read data session exists"
                )
            yield storage_session
        else:
            connection_id = self.storage_api.connect(
                self.access_token, exclusive=exclusive, write=write, create=create
            )
            if connection_id is not None:
                try:
                    storage_session = StorageSession(self.storage_api, connection_id)
                    self._current_data_session = (storage_session, exclusive, write)
                    yield storage_session
                    self._current_data_session = None
                    self.storage_api.disconnect(connection_id, rollback=False)
                except Exception:
                    self._current_data_session = None
                    self.storage_api.disconnect(connection_id, rollback=True)
                    raise
            else:
                raise RuntimeError("Failed to establish a data session.")

    @contextmanager
    def schema(self):
        connection_id = self.storage_api.connect(
            self.access_token, exclusive=True, write=True, create=True
        )
        try:
            yield SchemaSession(self.storage_api, connection_id)
            self.storage_api.disconnect(connection_id, rollback=False)
        except Exception:
            self.storage_api.disconnect(connection_id, rollback=True)
            raise

    def start_session(self, exclusive=None, write=False, create=False):
        connection_id = self.storage_api.connect(
            self.access_token, exclusive=exclusive, write=write, create=create
        )
        return StorageSession(self.storage_api, connection_id)

    def end_session(self, storage_session, rollback=False):
        self.storage_api.disconnect(storage_session._connection_id, rollback=rollback)


class SchemaSession:
    @classmethod
    def find_schema(cls, name, version_selection=None):
        module = importlib.import_module(name)

        schemas_to_collections = getattr(module, "_schemas_to_collections", None)
        if schemas_to_collections is None:
            schemas_to_collections = {}
            index = 0
            for schema_declaration in module.schemas:
                unknown_items = set(schema_declaration) - {"version", "schema"}
                if unknown_items:
                    raise ValueError(
                        f"invalid item(s) in schema {module.__name__} (list index={index}): {','.join(unknown_items)}"
                    )
                version = schema_declaration.get("version")
                if version is None:
                    raise ValueError(
                        f"version missing in schema {module.__name__} (list index={index})"
                    )
                sversion = version.split(".")
                if len(sversion) != 3:
                    raise ValueError(
                        f"invalid version in schema {module.__name__} (list index={index}): {version}"
                    )
                schema_content = schema_declaration.get("schema")
                if schema_content is None:
                    raise ValueError(
                        f"schema missing in schema {module.__name__} (list index={index})"
                    )
                try:
                    collections = cls._parse_schema_content(schema_content)
                except Exception as e:
                    raise ValueError(
                        f"invalid schema definition {module.__name__} (list index={index})"
                    ) from e
                schema_to_collections = {
                    "name": module.__name__,
                    "version": version,
                    "collections": collections,
                }
                if version in schemas_to_collections:
                    raise ValueError(
                        f"two schemas with version {version} defined in {module.__name__}"
                    )
                schemas_to_collections[version] = schema_to_collections
                if None not in schemas_to_collections:
                    schemas_to_collections[None] = schema_to_collections
                short_version = ".".join(sversion[:2])
                if short_version not in schemas_to_collections:
                    schemas_to_collections[short_version] = schema_to_collections
                index += 1
            module._schemas_to_collections = schemas_to_collections
        return schemas_to_collections.get(version_selection)

    @classmethod
    def _parse_field(cls, name, definition):
        error = True
        if isinstance(definition, type | types.GenericAlias):
            type_str = type_to_str(definition)
            kwargs = {}
            error = False
        elif isinstance(definition, str):
            type_str = definition
            kwargs = {}
            error = False
        elif isinstance(definition, list) and len(definition) == 2:
            if isinstance(definition[1], dict):
                if isinstance(definition[0], type | types.GenericAlias):
                    type_str = type_to_str(definition[0])
                    kwargs = definition[1]
                    error = False
                elif isinstance(definition[0], str):
                    type_str = definition[0]
                    kwargs = definition[1]
                    error = False
        if error:
            raise ValueError(f'invalid definition for field "{name}"')
        return [type_str, kwargs]

    @classmethod
    def _parse_schema_content(cls, schema_content):
        if not isinstance(schema_content, dict):
            raise ValueError("schema must be a dict")
        collections = {}
        for k, v in schema_content.items():
            if isinstance(v, dict):
                fields = {kk: cls._parse_field(kk, vv) for kk, vv in v.items()}
                for name, d in fields.items():
                    _, kwargs = d
                    if name == Storage.default_field:
                        raise ValueError(f'invalid field name: "{name}')
                    if "primary_key" in kwargs:
                        raise ValueError(
                            f'primary key not allowed for unique document "{k}" in field "{name}"'
                        )
                fields[Storage.default_field] = ["str", {"primary_key": True}]
                collections[k] = fields
            elif isinstance(v, list) and len(v) == 1 and isinstance(v[0], dict):
                fields = {kk: cls._parse_field(kk, vv) for kk, vv in v[0].items()}
                primary = False
                for name, d in fields.items():
                    type, kwargs = d
                    if name == Storage.default_field:
                        raise ValueError(f'invalid field name: "{name}')
                    if "primary_key" in kwargs:
                        primary = True
                if not primary:
                    raise ValueError(f'no primary key defined for collection "{k}"')
                collections[k] = fields
            else:
                type_str, kwargs = cls._parse_field(k, v)
                fields = collections.setdefault(Storage.default_collection, {})
                if Storage.default_field not in fields:
                    fields[Storage.default_field] = ["str", {"primary_key": True}]
                fields[k] = [type_str, kwargs]
        return collections

    def __init__(self, server, connection_id):
        super().__setattr__("_storage_api", server)
        super().__setattr__("_connection_id", connection_id)

    def add_schema(self, name, version=None):
        schema_to_collections = self.find_schema(name, version)
        if not schema_to_collections:
            raise ValueError(f"cannot find schema {name} with version {version}")
        self._storage_api.add_schema_collections(
            self._connection_id, schema_to_collections
        )

    def add_collection(self, name, primary_key):
        # Make primary_key json compatible
        if isinstance(primary_key, dict):
            primary_key = dict(
                (k, (v if isinstance(v, str) else type_to_str(v)))
                for k, v in primary_key.items()
            )
        self._storage_api.add_collection(self._connection_id, name, primary_key)

    def add_field(
        self, collection_name, field_name, field_type, description=None, index=False
    ):
        if isinstance(field_type, type | types.GenericAlias):
            field_type = type_to_str(field_type)
        self._storage_api.add_field(
            self._connection_id,
            collection_name,
            field_name,
            field_type,
            description,
            index,
        )

    def remove_field(self, collection_name, field_name):
        """
        Removes a specified field from a collection in the storage system.

        Args:
            collection_name (str): The name of the collection containing
                                   the field.
            field_name (str): The name of the field to be removed.

        This method delegates the operation to the storage API, ensuring
        the field is removed from the specified collection within the
        active connection.

        """
        self._storage_api.remove_field(self._connection_id, collection_name, field_name)

    def clear_database(self):
        return self._storage_api.clear_database(self._connection_id)

    @contextmanager
    def data(self):
        yield StorageSession(self._storage_api, self._connection_id)


class StorageSession:
    def __init__(self, storage_api, connection_id, path=None):
        super().__setattr__("_storage_api", storage_api)
        super().__setattr__("_connection_id", connection_id)
        super().__setattr__("_path", path or [])

    def __getitem__(self, key):
        return self.__class__(
            self._storage_api, self._connection_id, self._path + [key]
        )

    def __getattr__(self, key):
        return self[key]

    def __setitem__(self, key, value):
        self._storage_api.set(self._connection_id, self._path + [key], value)

    def __setattr__(self, key, value):
        self[key] = value

    def __delitem__(self, key):
        self._storage_api.delete(self._connection_id, self._path + [key])

    def __delattr__(self, key):
        del self[key]

    def primary_key(self):
        return self._storage_api.primary_key(self._connection_id, self._path)

    def set(self, value):
        self._storage_api.set(self._connection_id, self._path, value)

    def update(self, value):
        self._storage_api.update(self._connection_id, self._path, value)

    def get(self, default=None, fields=None, as_list=False, distinct=False):
        return self._storage_api.get(
            self._connection_id,
            self._path,
            default=default,
            fields=fields,
            as_list=as_list,
            distinct=distinct,
        )

    def count(self, query=None):
        return self._storage_api.count(self._connection_id, self._path, query=query)

    def append(self, value):
        return self._storage_api.append(self._connection_id, self._path, value)

    def distinct_values(self, field):
        return self._storage_api.distinct_values(self._connection_id, self._path, field)

    def search(self, query=None, fields=None, as_list=None, distinct=False, **kwargs):
        if kwargs and query:
            raise ValueError("Cannot combine query and equality research")
        if kwargs:
            query = " AND ".join(f'{{{k}}}=="{v}"' for k, v in kwargs.items())
        if isinstance(fields, tuple):
            fields = list(fields)
        return self._storage_api.search(
            self._connection_id,
            self._path,
            query,
            fields=fields,
            as_list=as_list,
            distinct=distinct,
        )

    def search_and_delete(self, query=None, **kwargs):
        if kwargs and query:
            raise ValueError("Cannot combine query and equality research")
        if kwargs:
            query = " AND ".join(f'{{{k}}}=="{v}"' for k, v in kwargs.items())
        return self._storage_api.search_and_delete(
            self._connection_id, self._path, query
        )

    def has_collection(self, collection):
        return self._storage_api.has_collection(
            self._connection_id, self._path, collection
        )

    def collection_names(self):
        return self._storage_api.collection_names(
            self._connection_id,
            self._path,
        )

    def keys(self):
        return self._storage_api.keys(
            self._connection_id,
            self._path,
        )



# import os
# import signal
import subprocess
import sys
import time
from datetime import datetime
from tempfile import NamedTemporaryFile

import pytest

from populse_db import Storage
from populse_db.storage import SchemaSession

snapshots = [
    {
        "orientation": "coronal",
        "top": [0.0, 0.0],
        "size": [872.0, 890.0],
        "dataset": "database_brainvisa",
        "software": "morphologist",
        "time_point": "M0",
        "data_type": "greywhite",
        "side": "both",
        "image": "/home/yann/dev/snapcheck/database_brainvisa/snapshots/morphologist/M0/greywhite/snapshot_greywhite_0001292COG_M0.png",
        "subject": "0001292COG",
    },
    {
        "orientation": "coronal",
        "top": [0.0, 0.0],
        "size": [872.0, 887.0],
        "dataset": "database_brainvisa",
        "software": "morphologist",
        "time_point": "M0",
        "data_type": "greywhite",
        "side": "both",
        "image": "/home/yann/dev/snapcheck/database_brainvisa/snapshots/morphologist/M0/greywhite/snapshot_greywhite_0001017COG_M0.png",
        "subject": "0001017COG",
    },
    {
        "orientation": "coronal",
        "top": [872.0, 0.0],
        "size": [872.0, 890.0],
        "dataset": "database_brainvisa",
        "software": "morphologist",
        "time_point": "M0",
        "data_type": "greywhite",
        "side": "both",
        "image": "/home/yann/dev/snapcheck/database_brainvisa/snapshots/morphologist/M0/greywhite/snapshot_greywhite_0001235COG_M0.png",
        "subject": "0001235COG",
    },
    {
        "orientation": "axial",
        "top": [1.0, 0.0],
        "size": [2.0, 3.0],
        "dataset": "test",
        "software": "none",
        "time_point": "first",
        "data_type": "void",
        "image": "/somewhere/something.png",
        "subject": "john doe",
    },
]


def test_storage_schema():
    snapshot_1_0_1 = {
        "name": "populse_db.test.schema.snapshot",
        "version": "1.0.1",
        "collections": {
            "snapshots": {
                "data_type": ["str", {"primary_key": True}],
                "execution": ["str", {"index": True}],
                "image": ["str", {}],
                "side": ["str", {}],
                "size": ["list[float]", {}],
                "subject": ["str", {"primary_key": True}],
                "time_point": ["str", {"primary_key": True}],
                "top": ["list[int]", {}],
            }
        },
    }
    snapshot_1_0_0 = {
        "name": "populse_db.test.schema.snapshot",
        "version": "1.0.0",
        "collections": {
            "snapshots": {
                "data_type": ["str", {"primary_key": True}],
                "image": ["str", {}],
                "side": ["str", {}],
                "size": ["list[float]", {}],
                "subject": ["str", {"primary_key": True}],
                "time_point": ["str", {"primary_key": True}],
                "top": ["list[int]", {}],
            }
        },
    }
    schema = SchemaSession.find_schema("populse_db.test.schema.snapshot")
    assert schema == snapshot_1_0_1
    schema = SchemaSession.find_schema("populse_db.test.schema.snapshot", "1.0.1")
    assert schema == snapshot_1_0_1
    schema = SchemaSession.find_schema("populse_db.test.schema.snapshot", "1.0.0")
    assert schema == snapshot_1_0_0
    schema = SchemaSession.find_schema("populse_db.test.schema.snapshot", "1.0")
    assert schema == snapshot_1_0_1
    assert SchemaSession.find_schema("populse_db.test.schema.snapshot", "1") is None
    assert SchemaSession.find_schema("populse_db.test.schema.snapshot", "1.1") is None

    import populse_db.test.test_schema

    assert set(populse_db.test.schema.snapshot._schemas_to_collections) == {
        "1.0.1",
        "1.0.0",
        "1.0",
        None,
    }

    with pytest.raises(ModuleNotFoundError):
        SchemaSession.find_schema("populse_db.test.schema.non_existing")


def run_storage_tests(store):
    with store.schema() as schema:
        schema.add_collection("test_collection_1", "primary_key")
        schema.add_collection("test_collection_1", "primary_key")
        with pytest.raises(ValueError):
            schema.add_collection("test_collection_1", "another_key")
        schema.add_collection(
            "test_collection_2", {"primary_key_1": "str", "primary_key_2": int}
        )
        schema.add_collection(
            "test_collection_2", {"primary_key_1": "str", "primary_key_2": int}
        )
        with pytest.raises(ValueError):
            schema.add_collection(
                "test_collection_2", ("primary_key_1", "primary_key_2")
            )
        with pytest.raises(ValueError):
            schema.add_collection(
                "test_collection_3", {"primary_key_1": "str", "primary_key_2": "what ?"}
            )

        schema.add_field("test_collection_1", "test", list[int])
        schema.add_field("test_collection_1", "test", list[int])
        schema.add_field("test_collection_1", "test", "list[int]")
        with pytest.raises(ValueError):
            schema.add_field("test_collection_1", "test", list[float])
        with pytest.raises(ValueError):
            schema.add_field("test_collection_1", "test_2", "what ?")
        schema.add_schema("populse_db.test.schema.dataset")
        schema.add_schema("populse_db.test.schema.snapshot")
        schema.add_schema("populse_db.test.schema.processing")

        schema.add_collection("test_update", "key")
        schema.add_field("test_update", "f1", str)
        schema.add_field("test_update", "f2", str)
        schema.add_field("test_update", "f3", str)
        schema.add_field("test_update", "dict", dict)
        schema.add_field("test_update", "yes", bool)
        schema.add_field("test_update", "no", bool)

        with schema.data() as d:
            d.a_name = "a value"

    with store.data(write=True) as d:
        now = datetime.now()

        # Set a global value
        d.last_update = now
        d.a_dict = {"one": 1}

        # Read a global value
        assert d.last_update.get() == now
        assert d.a_dict.get() == {"one": 1}

        assert d.does_not_exist.get() is None

        # Modify global value
        d.a_dict["two"] = 2
        assert d.a_dict.get() == {"one": 1, "two": 2}

        del d.a_dict["two"]
        assert d.a_dict.get() == {"one": 1}

        # Set single document fields
        d.dataset.directory = "/somewhere"
        d.dataset.schema.set("BIDS")
        # Following field is not in schema
        my_data = {
            "creation_date": now,
            "manager": "me",
            "my_list": ["zero", "one", "two"],
        }
        d.dataset.my_data = my_data
        assert d.dataset.directory.get() == "/somewhere"
        assert d.dataset.get() == {
            "directory": "/somewhere",
            "schema": "BIDS",
            "my_data": my_data,
        }
        assert d.dataset.my_data.my_list[2].get() == "two"

        # Get primary key of a collection
        assert d.test_collection_1.primary_key() == ["primary_key"]
        assert d.test_collection_2.primary_key() == ["primary_key_1", "primary_key_2"]
        with pytest.raises(ValueError):
            d.primary_key()
        with pytest.raises(ValueError):
            d.test_collection_1.any_field.primary_key()

        # Set values not in schema
        d.dataset.my_data.new_value = now
        assert d.dataset.my_data.new_value.get() == now
        d.dataset.my_data.my_list.append("three")
        assert d.dataset.my_data.my_list[3].get() == "three"
        d.dataset.my_data.my_list[3] = "last"
        assert d.dataset.my_data.my_list[3].get() == "last"

        # Update a document
        d.test_update["test"] = {
            "yes": True,
            "no": False,
            "f1": "f1",
            "f3": "value",
            "dict": {"one": 1, "three": 4},
        }
        d.test_update.test.update(
            {
                "f2": "f2",
                "f3": "f3",
            }
        )
        d.test_update.test.dict.update(
            {
                "two": 2,
                "three": 3,
            }
        )

        assert d.test_update.test.get() == {
            "yes": True,
            "no": False,
            "key": "test",
            "f1": "f1",
            "f2": "f2",
            "f3": "f3",
            "dict": {"one": 1, "two": 2, "three": 3},
        }

        d.test_update.test.update({})
        assert d.test_update.test.get() == {
            "yes": True,
            "no": False,
            "key": "test",
            "f1": "f1",
            "f2": "f2",
            "f3": "f3",
            "dict": {"one": 1, "two": 2, "three": 3},
        }

        # Delete data
        del d.test_update.test.dict.two
        del d.test_update.test.f2
        assert d.test_update.test.get() == {
            "yes": True,
            "no": False,
            "key": "test",
            "f1": "f1",
            "f2": None,
            "f3": "f3",
            "dict": {"one": 1, "three": 3},
        }
        assert d.test_update.test.yes.get() is True
        assert d.test_update.test.no.get() is False

        del d.test_update.test
        assert d.test_update.get() == []

        with pytest.raises(ValueError):
            d.update({})
        with pytest.raises(ValueError):
            d.test_update.update({})

        # Adds many documents in a collection
        for snapshot in snapshots:
            d.snapshots.append(snapshot)

        # Select one document from a collection (ignore fields with None value)
        assert {
            k: v
            for k, v in d.snapshots["0001292COG", "M0", "greywhite"].get().items()
            if v is not None
        } == snapshots[0]

        # Select one document field from a collection
        assert (
            d.snapshots["0001292COG", "M0", "greywhite"].image.get()
            == snapshots[0]["image"]
        )

        # Select all documents from a collection (ignore fields with None value)
        assert [
            {k: v for k, v in d.items() if v is not None} for d in d.snapshots.get()
        ] == snapshots

        # Set a full document content
        modified_snapshot = snapshots[0].copy()
        modified_snapshot["software"] = "something"
        del modified_snapshot["dataset"]
        d.snapshots["0001292COG", "M0", "greywhite"] = modified_snapshot
        assert {
            k: v
            for k, v in d.snapshots["0001292COG", "M0", "greywhite"].get().items()
            if v is not None
        } == modified_snapshot

        # Set a whole collection
        d["snapshots"] = []
        assert d.snapshots.get() == []
        d.snapshots = snapshots
        assert [
            {k: v for k, v in d.items() if v is not None} for d in d.snapshots.get()
        ] == snapshots

        # Search in a collection
        assert [
            {k: v for k, v in d.items() if v is not None}
            for d in d.snapshots.search('subject LIKE "%7%"')
        ] == snapshots[1:2]

        assert [
            {k: v for k, v in doc.items() if v is not None}
            for doc in d.snapshots.search(image="/somewhere/something.png")
        ] == snapshots[3:4]

        assert d.snapshots.search(
            time_point="M0", as_list=True, fields=["data_type"], distinct=True
        ) == [["greywhite"]]

        # Count elements
        assert d.snapshots.count() == 4
        assert d.snapshots.count('image LIKE "/home/yann%"') == 3

        # Find all unique values
        assert set(d.snapshots.distinct_values("data_type")) == {"greywhite", "void"}
        assert set(
            row[0]
            for row in d.snapshots.get(
                fields=["data_type"], as_list=True, distinct=True
            )
        ) == {"greywhite", "void"}

        # Get non existent data
        assert d.nothing.get() is None
        assert d.nothing.get("default") == "default"
        assert d.nothing.x.get() is None
        assert d.nothing.x.get("default") == "default"
        assert d.nothing.x.y.get() is None
        assert d.nothing.x.y.get("default") == "default"
        assert d.snapshots["0001292COG", "bad", "greywhite"].get() is None
        assert d.snapshots["0001292COG", "bad", "greywhite"].get("default") == "default"
        assert d.snapshots["0001292COG", "M0", "greywhite"].nothing.get() is None
        assert (
            d.snapshots["0001292COG", "M0", "greywhite"].nothing.get("default")
            == "default"
        )

        # Search and delete
        d.snapshots.search_and_delete('image LIKE "/home/yann%"')
        assert len(d.snapshots.get()) == 1

        # Check schema
        assert d.has_collection("unknown_collection") is False
        assert d.has_collection("snapshots") is True
        with pytest.raises(ValueError):
            d.snapshots.has_collection("snapshots")
        assert d.collection_names() == [
            "test_collection_1",
            "test_collection_2",
            "dataset",
            "metadata",
            "snapshots",
            "execution",
            "test_update",
        ]
        with pytest.raises(ValueError):
            d.snapshots.collection_names()
        assert set(d.snapshots.keys()) == {
            "subject",
            "time_point",
            "image",
            "top",
            "size",
            "execution",
            "data_type",
            "side",
        }
        with pytest.raises(ValueError):
            d.keys()
        with pytest.raises(ValueError):
            d.snapshots.subject.keys()

        # Test reentrant data session
        with store.data(write=False) as d2:
            assert d2.last_update.get() == now
        with store.data(write=True) as data:
            data.anything = "something"
            assert data.anything.get() == "something"
            data.anything = 42
            assert data.anything.get() == 42
            data.anything = {}
            assert data.anything.get() == {}
            data.test_collection_1.value = now
            assert data.test_collection_1.value.get() == now
            data.test_collection_1.value = 42
            assert data.test_collection_1.value.get() == 42
            data.test_collection_1.value = "toto"
            assert data.test_collection_1.value.get() == "toto"

    # Test field removal
    with store.schema() as schema:
        schema.add_collection("test_collection_3", "primary_key")
        schema.add_field("test_collection_3", "a", str)
        schema.add_field("test_collection_3", "b", str)
        schema.add_field("test_collection_3", "x", str)
        schema.add_field("test_collection_3", "c", str)
    with store.data(write=True) as d:
        d.test_collection_3.key = {"a": "a", "b": "b", "c": "c", "x": "x"}
        assert d.test_collection_3.key.get() == {
            "primary_key": "key",
            "a": "a",
            "b": "b",
            "c": "c",
            "x": "x",
        }
        assert set(d.test_collection_3.keys()) == {"primary_key", "a", "b", "x", "c"}
    with store.schema() as schema:
        schema.remove_field("test_collection_3", "x")
    with store.data(write=False) as d:
        assert d.test_collection_3.key.get() == {
            "primary_key": "key",
            "a": "a",
            "b": "b",
            "c": "c",
        }
        assert set(d.test_collection_3.keys()) == {"primary_key", "a", "b", "c"}

    # Test read only session
    with store.data(write=False) as d:
        with pytest.raises(PermissionError):
            d.last_update = now

        with pytest.raises(PermissionError):
            for snapshot in snapshots:
                d.snapshots.append(snapshot)

        # Test reentrant data session
        with store.data(write=False) as d2:
            assert d2.anything.get() == {}
        with pytest.raises(RuntimeError):
            with store.data(write=True) as d2:
                d2.anything = "something"


def test_storage():
    store = Storage("/tmp/i_do_not_exist")
    with pytest.raises(RuntimeError):
        with store.data() as d:
            pass

    with NamedTemporaryFile(delete=True) as tmp:
        tmp.close()
        store = Storage(tmp.name)
        run_storage_tests(store)


def test_storage_server():
    pytest.importorskip("fastapi")
    pytest.importorskip("uvicorn")
    pytest.importorskip("tblib")

    with NamedTemporaryFile(delete=True) as tmp:
        tmp_path = tmp.name
        tmp.close()
        cmd = [sys.executable, "-m", "populse_db.server", tmp_path]
        server = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        try:
            time.sleep(5)
            store = Storage(tmp_path)
            run_storage_tests(store)

        except Exception as e:
            print("Error during test execution:", e)
            raise

        finally:
            server.terminate()


#        finally:
#            os.kill(server.pid, signal.SIGTERM)



import argparse
import os
import subprocess
import sys
from pathlib import Path

parser = argparse.ArgumentParser(
    prog="Test populse_db module", description="Run populse_db test suite"
)
parser.add_argument(
    "--html",
    action="store",
    default=None,
    metavar="DIRECTORY",
    help="Create tests and coverage HTML reports in given "
    "directory. Entrypoints are test.html and index.html",
)
parser.add_argument(
    "-v", "--verbose", action="count", default=0, help="Increase verbosity"
)
parser.add_argument(
    "-s", action="store_true", default=None, help="Do not capture stdout nor stderr"
)
parser.add_argument(
    "-x",
    "--exitfirst",
    action="store_true",
    help="Exit instantly on first error or failed test",
)
parser.add_argument(
    "-k",
    metavar="EXPRESSION",
    dest="keyword",
    default=None,
    action="store",
    help="Only run tests which match the given substring "
    "expression. An expression is a Python evaluatable "
    "expression where all names are substring-matched "
    "against test names and their parent classes. "
    "Example: -k 'test_method or test_other' matches "
    "all test functions and classes whose name contains "
    "'test_method' or 'test_other', while -k 'not "
    "test_method'vmatches those that don't contain "
    "'test_method' in their names. -k 'not test_method "
    "and not test_other' will eliminate the matches. "
    "Additionally keywords are matched to classes and "
    "functions containing extra names in their "
    "'extra_keyword_matches' set, as well as functions "
    "which have names assigned directly to them. The "
    "matching is case-insensitive.",
)

args = parser.parse_args()
pytest_command = [sys.executable, "-m", "pytest"]
if args.verbose:
    pytest_command.append("-" + "v" * args.verbose)
if args.exitfirst:
    pytest_command.append("-x")
if args.keyword:
    pytest_command += ["-k", args.keyword]
if args.s:
    pytest_command += ["-s"]
module_src = Path(__file__).parent.parent
if args.html:
    # Path(args.html).mkdir(exist_ok=True)
    pytest_command += ["--cov=populse_db", f"--html={args.html}/tests.html"]
    coverage_command = [sys.executable, "-m", "coverage", "html", "-d", args.html]
    env = os.environ.copy()
    print(" ".join(f"'{i}'" for i in pytest_command))
    subprocess.check_call(pytest_command, env=env, cwd=module_src)
    print(" ".join(f"'{i}'" for i in coverage_command))
    subprocess.check_call(coverage_command, cwd=module_src)
else:
    print(" ".join(f"'{i}'" for i in pytest_command))
    subprocess.check_call(pytest_command, cwd=module_src)



import importlib.metadata
import os
import re
import threading
from contextlib import contextmanager
from urllib.parse import urlparse

try:
    __version__ = importlib.metadata.__version__ = importlib.metadata.version(
        "populse_db"
    )
except importlib.metadata.PackageNotFoundError:
    __version__ = None


class Database:
    """Entrypoint of populse_db for creating :any:`DatabaseSession` object given
    an URL that identify the underlying database engine.

    Creating a :any:`Database` doesn't connect to the database engine. It just
    parses the URL (ensuring that it corresponds to a valid engine) and stores
    it. The :any:`Database` is a Python context manager that must be using a
    ``with`` statement that connects to the database and creates a database
    specific object that implements the API of :any:`DatabaseSession`.

    Example::

        from populse_db import Database

        db = Database('sqlite:///tmp/populse_db.sqlite')
        with db as dbs:
            dbs.add_collection('my_collection', primary_key='id')
            dbs['my_collection']['my_document'] = {
                'a key': 'a value',
                'another key': 'another value'
            }

    Database modification within a ``with`` statement is done in a transaction.
    If an exception occurs before the end of the `with`, a rollback is done and
    the database is not modified. By default, this transaction allows to read
    the database in parallel. The first database modification will acquire an
    exclusive access to the database. It is possible to request an exclusive
    session (preventing any other access to the database by other processes
    or threads) by using the `exclusive` property::

        from populse_db import Database

        db = Database('sqlite:///tmp/populse_db.sqlite')
        with db.exclusive as dbs:
            # All other access to the database are blocked
            # until the end of this "with" statement.

    :any:``Database`` is a reusable context manager. It means that it is allowed
    to use it in several consecutive ``with``::

        from populse_db import Database

        db = Database('sqlite:///tmp/populse_db.sqlite')
        with db as dbs:
            dbs.add_collection('my collection')

        with db as dbs:
            dbs['my collection']['my document'] = {}

    :any:`Database` is a reentrant context manager. It is possible to use a
    ``with`` statement within another ``with`` that already use the same
    :any:`Database`. In that case, the same context is returned and the
    transaction is not ended in any of the inner context, it is terminated
    when the outer context exits::

        from populse_db import Database

        db = Database('sqlite:///tmp/populse_db.sqlite')
        with db as dbs1:
            # Connection to the database and transaction starts here
            dbs.add_collection('my collection')
            with db as dbs2:
                # here: dbs1 is dbs2 == True
                # No new connection to the database is done
                dbs['my collection']['my document'] = {}
            # After the inner with, connection to the database is not
            # closed and transaction is ongoing (no commit nor rollback
            # is done)
        # After the end of the outer with, transaction is terminated and
        # connection to the database is closed.

    On a multithreaded application, each thread using :any:`Database` context
    manager gets its own :any:`DatabaseSession` instance with its own database
    connection.

    methods:
        - __enter__: Creates a :any:`DatabaseSession` instance
        - __exit__: Release resource used by the :any:`DatabaseSession`
    """

    def __init__(self, database_url, timeout=None, create=False, echo_sql=None):
        """Creates a :any:`Database` instance.

        :param database_url: URL defining database engine and its parameters. The
            engine URL must have the following pattern:
            dialect://user:password@host/dbname[?key=value..].
            To date dialect can only be ``sqlite`` but ``postgresql`` is planned.

            Examples:
                    - /somewhere/a_file.popdb
                    - ``sqlite:///foo.db``
                    - ``postgresql://scott:tiger@localhost/test``
        """

        self.thread_local = threading.local()

        # Remove the drive letter for Windows paths
        if os.name == "nt" and re.match(r"^[a-zA-Z]:\\", database_url):
            database_url = database_url[2:]

        self.url = urlparse(database_url)
        self.create = create
        self.echo_sql = echo_sql

        if timeout:
            self.timeout = int(timeout)

        else:
            self.timeout = None

        if self.url.scheme in ("", "sqlite"):
            # sqlite3 module (used in .engine.sqlite) is optional
            # because it does not exist in Pyodide distribution (i.e.
            # PyScript).
            from .engine.sqlite import create_sqlite_session_factory
            self.session_factory = create_sqlite_session_factory(self.url)

        else:
            raise ValueError(f"Invalid database type in database URL: {database_url}")

    def session(self, exclusive=False, create=None):
        if create is None:
            create = self.create
        return self.session_factory(
            exclusive=exclusive,
            timeout=self.timeout,
            echo_sql=self.echo_sql,
            create=create,
        )

    def begin_session(self, exclusive, create=None):
        if create is None:
            create = self.create
        session_depth = getattr(self.thread_local, "populse_db", None)
        if session_depth is None:
            session = self.session(exclusive=exclusive, create=create)
            depth = 0
        else:
            session, depth = session_depth
        depth += 1
        self.thread_local.populse_db = (session, depth)
        return session

    def end_session(self, rollback):
        session, depth = self.thread_local.populse_db
        depth -= 1
        if depth == 0:
            session.close(rollback=rollback)
            del self.thread_local.populse_db
        else:
            self.thread_local.populse_db = (session, depth)

    def __enter__(self):
        """
        Return a DatabaseSession instance for using the database. This is
        supposed to be called using a "with" statement:

        with database as session:
           session.add_document(...)

        Therefore __exit__ must be called to get rid of the session.
        When called recursively, the underlying database session returned
        is the same. The commit/rollback of the session is done only by the
        outermost __enter__/__exit__ pair (i.e. by the outermost with
        statement).
        """
        return self.begin_session(exclusive=False, create=self.create)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_session(rollback=(exc_type is not None))

    @property
    @contextmanager
    def exclusive(self, create=None):
        if create is None:
            create = self.create
        try:
            session = self.begin_session(exclusive=True, create=create)
            yield session
            self.end_session(rollback=False)
        except Exception:
            self.end_session(rollback=True)
            raise


# Import here to allow the following import in external
# modules:
from .database import json_decode, json_encode  # noqa: F401, E402
from .storage import Storage  # noqa: F401, E402



import os
import shutil
import tempfile
import unittest
from datetime import date, datetime, time

from populse_db import Database
from populse_db.database import check_value_type, populse_db_table

# from populse_db.engine.sqlite import SQLiteSession
from populse_db.filter import FilterToSQL, literal_parser


class TestsSQLiteInMemory(unittest.TestCase):
    def test_add_get_document(self):
        now = datetime.now()
        db = Database("sqlite://:memory:", create=True)
        with db as dbs:
            dbs.add_collection("test", "index")
            base_doc = {
                "string": "string",
                "int": 1,
                "float": 1.4,
                "boolean": True,
                "datetime": now,
                "date": now.date(),
                "time": now.time(),
                "dict": {
                    "string": "string",
                    "int": 1,
                    "float": 1.4,
                    "boolean": True,
                },
            }
            doc = base_doc.copy()
            for k, v in base_doc.items():
                lk = f"list_{k}"
                doc[lk] = [v]
            doc["index"] = "test"
            dbs.add_document("test", doc)
            stored_doc = dbs.get_document("test", "test")
            self.maxDiff = None
            self.assertEqual(doc, stored_doc)


def create_test_case(**database_creation_parameters):
    class TestDatabaseMethods(unittest.TestCase):
        """
        Class executing the unit tests of populse_db
        """

        def setUp(self):
            """
            Called before every unit test
            Creates a temporary folder containing the database file that will be used for the test
            """
            self.database_creation_parameters = database_creation_parameters.copy()
            if "database_url" not in self.database_creation_parameters:
                self.temp_folder = tempfile.mkdtemp(prefix="populse_db")
                path = os.path.join(self.temp_folder, "test.db")
                self.database_creation_parameters["database_url"] = "sqlite://" + path
            else:
                self.temp_folder = None
            self.database_url = self.database_creation_parameters["database_url"]

        def tearDown(self):
            """
            Called after every unit test
            Deletes the temporary folder created for the test
            """
            if self.temp_folder:
                shutil.rmtree(self.temp_folder)
                del self.database_creation_parameters["database_url"]
            self.temp_folder = None

        def create_database(self, clear=True, create=True, echo_sql=None):
            """
            Opens the database
            :param clear: Bool to know if the database must be cleared
            """

            try:
                db = Database(
                    **self.database_creation_parameters,
                    create=create,
                    echo_sql=echo_sql,
                )
            except Exception as e:
                if self.database_creation_parameters["database_url"].startswith(
                    "postgresql"
                ):
                    raise unittest.SkipTest(str(e)) from e
                raise
            except ImportError as e:
                if "psycopg2" in str(e) and self.database_creation_parameters[
                    "database_url"
                ].startswith("postgresql"):
                    raise unittest.SkipTest(str(e)) from e
                raise
            if clear:
                with db as dbs:
                    dbs.clear()
            return db

        def test_wrong_constructor_parameters(self):
            """
            Tests the parameters of the Database class constructor
            """
            # Testing with wrong engine
            self.assertRaises(
                ValueError, lambda: Database("engine://something").__enter__()
            )

        def test_add_field(self):
            """
            Tests the method adding a field
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Testing with a first field
                session.add_field(
                    "collection1", "PatientName", str, description="Name of the patient"
                )

                # Checking the field properties
                field = session.get_field("collection1", "PatientName")
                self.assertEqual(field["name"], "PatientName")
                self.assertEqual(field["type"], str)
                self.assertEqual(field["description"], "Name of the patient")
                self.assertEqual(field["collection"], "collection1")

                # Testing with a field that already exists
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.add_field(
                        "collection1", "PatientName", str, "Name of the patient"
                    ),
                )

                # Testing with several field types
                session.add_field("collection1", "BandWidth", float, None)
                session.add_field("collection1", "AcquisitionTime", time, None)
                session.add_field("collection1", "AcquisitionDate", datetime, None)
                session.add_field("collection1", "Dataset dimensions", list[int], None)
                session.add_field("collection1", "Boolean", bool, None)
                session.add_field("collection1", "Boolean list", list[bool], None)

                # Testing with close field names
                session.add_field("collection1", "Bits per voxel", int, "with space")
                session.add_field("collection1", "Bitspervoxel", int, "without space")
                with self.assertRaises(session.database_exceptions):
                    session.add_field("collection1", "bitspervoxel", int, "lower case")
                self.assertEqual(
                    session.get_field("collection1", "Bitspervoxel")["description"],
                    "without space",
                )
                self.assertEqual(
                    session.get_field("collection1", "Bits per voxel")["description"],
                    "with space",
                )
                with self.assertRaises(TypeError):
                    self.assertEqual(
                        session.get_field("collection1", "bitspervoxel")["description"],
                        "lower case",
                    )

                # Testing with wrong parameters
                self.assertRaises(
                    ValueError,
                    lambda: session.add_field(
                        "collection_not_existing", "Field", list[int], None
                    ),
                )
                self.assertRaises(
                    ValueError,
                    lambda: session.add_field(True, "Field", list[int], None),
                )
                self.assertRaises(
                    AttributeError,
                    lambda: session.add_field(
                        "collection1", "Patient Name", None, None
                    ),
                )

                # Testing that the document primary key field is taken
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.add_field("collection1", "name", str, None),
                )

                with self.assertRaises(ValueError):
                    session.remove_field("collection", "name")

        def test_add_fields(self):
            """
            Tests the method adding several fields
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1")

                # Adding several fields
                session.add_field("collection1", "First name", str, "")
                session.add_field("collection1", "Last name", str, "")
                collection_fields = list(session.get_fields_names("collection1"))
                self.assertEqual(len(collection_fields), 3)
                self.assertTrue("primary_key" in collection_fields)
                self.assertTrue("First name" in collection_fields)
                self.assertTrue("Last name" in collection_fields)

        def test_remove_field(self):
            """
            Tests the method removing a field
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("current", "name")

                # Adding fields
                session.add_field("current", "PatientName", str, "Name of the patient")
                session.add_field("current", "SequenceName", str, None)
                session.add_field("current", "Dataset dimensions", list[int], None)

                # Adding documents
                document = {}
                document["name"] = "document1"
                session.add_document("current", document)
                document = {
                    "name": "document2",
                    "PatientName": "Guerbet",
                    "SequenceName": "RARE",
                    "Dataset dimensions": [1, 2],
                }
                document["name"] = "document2"
                session.add_document("current", document)

                # Removing fields
                session.remove_field("current", "PatientName")
                session.remove_field("current", "Dataset dimensions")

                # Testing that the field does not exist anymore
                self.assertIsNone(session.get_field("current", "PatientName"))
                self.assertIsNone(session.get_field("current", "Dataset dimensions"))

                # Testing that the field values are removed
                with self.assertRaises(KeyError):
                    session["current"]["document1"]["PatientName"]
                with self.assertRaises(KeyError):
                    session["current"]["document2"]["PatientName"]
                self.assertIsNone(session["current"]["document1"]["SequenceName"])
                self.assertEqual(
                    session["current"]["document2"]["SequenceName"],
                    "RARE",
                )
                with self.assertRaises(KeyError):
                    session["current"]["document1"]["Dataset dimensions"]
                with self.assertRaises(KeyError):
                    session["current"]["document2"]["Dataset dimensions"]

                # /\ Deleting a list of fields is not yet implemented in populse_db 3.0! /\
                # Testing with list of fields
                # session.remove_field("current", ["SequenceName"])
                # self.assertIsNone(session.get_field("current", "SequenceName"))

                # Adding fields again
                session.add_field("current", "PatientName", str, "Name of the patient")
                # session.add_field("current", "SequenceName", str, None)
                session.add_field("current", "Dataset dimensions", list[int], None)

                # Testing with list of fields
                # session.remove_field("current", ["SequenceName", "PatientName"])
                # self.assertIsNone(session.get_field("current", "SequenceName"))
                # self.assertIsNone(session.get_field("current", "PatientName"))

                # Testing with a field not existing
                self.assertRaises(
                    ValueError,
                    lambda: session.remove_field("not_existing", "document1"),
                )
                self.assertRaises(
                    ValueError, lambda: session.remove_field(1, "NotExisting")
                )
                import sqlite3

                self.assertRaises(
                    sqlite3.OperationalError,
                    lambda: session.remove_field("current", "NotExisting"),
                )
                self.assertRaises(
                    sqlite3.OperationalError,
                    lambda: session.remove_field("current", "Dataset dimension"),
                )
                # /\ Deleting a list of fields is not yet implemented in populse_db 3.0! /\
                # self.assertRaises(
                #     ValueError,
                #     lambda: session.remove_field(
                #         "current", ["SequenceName", "PatientName", "Not_Existing"]
                #     ),
                # )

                # Testing with wrong parameters
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.remove_field("current", 1),
                )
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.remove_field("current", None),
                )

                # /\ Deleting a list of fields is not yet implemented in populse_db 3.0! /\
                # Removing list of fields with list type
                # session.add_field("current", "list1", list[int], None)
                # session.add_field("current", "list2", list[str], None)
                # session.remove_field("current", ["list1", "list2"])
                # self.assertIsNone(session.get_field("current", "list1"))
                # self.assertIsNone(session.get_field("current", "list2"))

        def test_get_field(self):
            """
            Tests the method giving the field row given a field
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding a field
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )

                # Testing that the field is returned if it exists
                self.assertIsNotNone(session.get_field("collection1", "PatientName"))

                # Testing that None is returned if the field does not exist
                self.assertIsNone(session.get_field("collection1", "Test"))

                # Testing that None is returned if the collection does not exist
                self.assertIsNone(
                    session.get_field("collection_not_existing", "PatientName")
                )

                # Testing that None is returned if both collection and field do not exist
                self.assertIsNone(session.get_field("collection_not_existing", "Test"))

        def test_get_fields(self):
            """
            Tests the method giving all fields rows, given a collection
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding a field
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )

                fields = session.get_fields("collection1")
                self.assertEqual(len(fields), 2)

                session.add_field(
                    "collection1", "SequenceName", str, "Name of the patient"
                )

                fields = session.get_fields("collection1")
                self.assertEqual(len(fields), 3)

                # Adding a second collection
                session.add_collection("collection2", "id")

                fields = session.get_fields("collection1")
                self.assertEqual(len(fields), 3)

                # Testing with a collection not existing
                self.assertEqual(
                    list(session.get_fields("collection_not_existing")), []
                )

        def test_set_values(self):
            """
            Tests the method setting several values of a document
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1")

                # Adding fields
                session.add_field("collection1", "SequenceName", str)
                session.add_field("collection1", "PatientName", str)
                session.add_field("collection1", "BandWidth", float)

                # Adding documents
                session["collection1"]["document1"] = {
                    "SequenceName": "Flash",
                    "PatientName": "Guerbet",
                    "BandWidth": 50000,
                }
                session["collection1"]["document2"] = {}

                # Adding values
                document1 = session["collection1"]["document1"]
                self.assertEqual(document1["SequenceName"], "Flash")
                self.assertEqual(document1["PatientName"], "Guerbet")
                self.assertEqual(document1["BandWidth"], 50000)

                # Setting all values
                values = {}
                values["PatientName"] = "Patient"
                values["BandWidth"] = 25000
                session.set_values("collection1", "document1", values)
                document1 = session["collection1"]["document1"]
                self.assertEqual(document1["SequenceName"], "Flash")
                self.assertEqual(document1["PatientName"], "Patient")
                self.assertEqual(document1["BandWidth"], 25000)

                # Testing that the primary_key cannot be set
                values = {"primary_key": "document3", "BandWidth": 25000}
                with self.assertRaises(ValueError):
                    session.set_values("collection1", "document1", values)

                # Trying with the field not existing
                values = {}
                values["PatientName"] = "Patient"
                values["BandWidth"] = 25000
                values["Field_not_existing"] = "value"
                session.set_values("collection1", "document1", values)

                # Trying with invalid values
                values = {}
                values["PatientName"] = 50
                values["BandWidth"] = 25000
                session.set_values("collection1", "document1", values)
                self.assertRaises(
                    AttributeError,
                    lambda: session.set_values("collection1", "document1", True),
                )

                # Trying with the collection not existing
                values = {}
                values["PatientName"] = "Guerbet"
                values["BandWidth"] = 25000
                self.assertRaises(
                    ValueError,
                    lambda: session.set_values(
                        "collection_not_existing", "document1", values
                    ),
                )

                # Trying with the document not existing
                values = {}
                values["PatientName"] = "Guerbet"
                values["BandWidth"] = 25000
                with self.assertRaises(ValueError):
                    session.set_values("collection1", "document_not_existing", values)

                # Testing with list values
                session.add_field("collection1", "list1", list[str])
                session.add_field("collection1", "list2", list[int])
                values = {}
                values["list1"] = ["a", "a", "a"]
                values["list2"] = [1, 1, 1]
                session.set_values("collection1", "document1", values)
                self.assertEqual(
                    session["collection1"]["document1"]["list1"], ["a", "a", "a"]
                )
                self.assertEqual(
                    session["collection1"]["document1"]["list2"], [1, 1, 1]
                )

        def test_get_field_names(self):
            """
            Tests the method removing a value
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding a field
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )

                fields = session.get_fields_names("collection1")
                self.assertEqual(len(fields), 2)
                self.assertTrue("name" in fields)
                self.assertTrue("PatientName" in fields)

                session.add_field(
                    "collection1", "SequenceName", str, "Name of the patient"
                )

                fields = session.get_fields_names("collection1")
                self.assertEqual(len(fields), 3)
                self.assertTrue("name" in fields)
                self.assertTrue("PatientName" in fields)
                self.assertTrue("SequenceName" in fields)

                # Adding a second collection
                session.add_collection("collection2", "id")

                fields = session.get_fields_names("collection1")
                self.assertEqual(len(fields), 3)
                self.assertTrue("name" in fields)
                self.assertTrue("PatientName" in fields)
                self.assertTrue("SequenceName" in fields)

                # Testing with a collection not existing
                self.assertEqual(
                    list(session.get_fields_names("collection_not_existing")), []
                )

        def test_get_value(self):
            """
            Tests the method giving the current value, given a document and a field
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding documents
                document = {}
                document["name"] = "document1"
                session.add_document("collection1", document)

                # Adding fields
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )
                session.add_field("collection1", "Dataset dimensions", list[int], None)
                session.add_field("collection1", "Bits per voxel", int, None)
                session.add_field("collection1", "Grids spacing", list[float], None)

                # Adding values
                document["PatientName"] = "test"
                document["Bits per voxel"] = 10
                document["Dataset dimensions"] = [3, 28, 28, 3]
                document["Grids spacing"] = [0.234375, 0.234375, 0.4]
                session["collection1"].add(document, replace=True)

                # Testing that the value is returned if it exists
                document1 = session["collection1"]["document1"]
                self.assertEqual(document1["PatientName"], "test")
                self.assertEqual(document1["Bits per voxel"], 10)
                self.assertEqual(document1["Dataset dimensions"], [3, 28, 28, 3])
                self.assertEqual(document1["Grids spacing"], [0.234375, 0.234375, 0.4])

        def test_check_type_value(self):
            """
            Tests the method checking the validity of incoming values
            """

            database = self.create_database()
            with database as _:
                self.assertTrue(check_value_type("string", str))
                self.assertFalse(check_value_type(1, str))
                self.assertTrue(check_value_type(None, str))
                self.assertTrue(check_value_type(1, int))
                self.assertTrue(check_value_type(1, float))
                self.assertTrue(check_value_type(1.5, float))
                self.assertTrue(check_value_type([1.5], list[float]))
                self.assertFalse(check_value_type(1.5, list[float]))
                self.assertFalse(check_value_type([1.5, "test"], list[float]))
                value = {}
                value["test1"] = 1
                value["test2"] = 2
                self.assertTrue(check_value_type(value, dict))
                value2 = {}
                value2["test3"] = 1
                value2["test4"] = 2
                self.assertTrue(check_value_type([value, value2], list[dict]))

        def test_add_value(self):
            """
            Tests the method adding a value
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding documents
                document = {}
                document["name"] = "document1"
                session.add_document("collection1", document)
                document = {}
                document["name"] = "document2"
                session.add_document("collection1", document)

                # Adding fields
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )
                session.add_field("collection1", "Bits per voxel", int, None)
                session.add_field("collection1", "BandWidth", float, None)
                session.add_field("collection1", "AcquisitionTime", time, None)
                session.add_field("collection1", "AcquisitionDate", datetime, None)
                session.add_field("collection1", "Dataset dimensions", list[int], None)
                session.add_field("collection1", "Grids spacing", list[float], None)
                session.add_field("collection1", "Boolean", bool, None)
                session.add_field("collection1", "Boolean list", list[bool], None)

                # Adding values
                collection1 = session["collection1"]

                document1 = collection1["document1"]
                document1["PatientName"] = "test"
                document1["Bits per voxel"] = 1
                document1["Dataset dimensions"] = [3, 28, 28, 3]
                document1["Boolean"] = True
                d = datetime(2014, 2, 11, 8, 5, 7)
                document1["AcquisitionDate"] = d
                t = datetime(2014, 2, 11, 0, 2, 2).time()
                document1["AcquisitionTime"] = t
                document1["BandWidth"] = 45
                collection1["document1"] = document1

                document2 = collection1["document2"]
                document2["Grids spacing"] = [0.234375, 0.234375, 0.4]
                document2["BandWidth"] = 35.5
                collection1["document2"] = document2

                # Testing that the values are actually added
                document1 = session["collection1"]["document1"]
                document2 = session["collection1"]["document2"]
                self.assertEqual(document1["PatientName"], "test")
                self.assertEqual(document2["BandWidth"], 35.5)
                self.assertEqual(document1["Bits per voxel"], 1)
                self.assertEqual(document1["BandWidth"], 45)
                self.assertEqual(document1["AcquisitionDate"], d)
                self.assertEqual(document1["AcquisitionTime"], t)
                self.assertEqual(document1["Dataset dimensions"], [3, 28, 28, 3])
                self.assertEqual(document2["Grids spacing"], [0.234375, 0.234375, 0.4])
                self.assertEqual(document1["Boolean"], True)

        def test_update_document(self):
            database = self.create_database()
            with database as session:
                session.add_collection("collection")
                collection = session["collection"]
                collection.add_field("status", str)
                collection.add_field("executable", dict)
                collection.add_field("execution_context", dict)

                doc = {
                    "primary_key": "doc",
                    "status": "submitted",
                    "executable": {"definition": "custom"},
                    "execution_context": {"tmp": "/tmp"},
                }
                collection["doc"] = doc
                self.assertEqual(collection["doc"], doc)

                collection.update_document(
                    "doc", {"status": "running", "other": "something"}
                )
                doc.update({"status": "running", "other": "something"})
                self.assertEqual(collection["doc"], doc)

        def test_update_document_without_field(self):
            database = self.create_database()
            with database as session:
                session.add_collection("collection")
                collection = session["collection"]

                doc = {
                    "primary_key": "doc",
                    "status": "submitted",
                    "executable": {"definition": "custom"},
                    "execution_context": {"tmp": "/tmp"},
                }
                collection["doc"] = doc
                self.assertEqual(collection["doc"], doc)

                collection.update_document(
                    "doc", {"status": "running", "other": "something"}
                )
                doc.update({"status": "running", "other": "something"})
                self.assertEqual(collection["doc"], doc)

        def test_get_document(self):
            """
            Tests the method giving the document row given a document
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding a document
                document = {}
                document["name"] = "document1"
                session.add_document("collection1", document)

                # Testing that a document is returned if it exists
                self.assertIsNotNone(session.get_document("collection1", "document1"))

                # Testing that None is returned if the document does not exist
                self.assertIsNone(session.get_document("collection1", "document3"))

                # Testing that None is returned if the collection does not exist
                self.assertIsNone(
                    session.get_document("collection_not_existing", "document1")
                )

                # Testing with wrong parameters
                self.assertIsNone(session.get_document(False, "document1"))
                self.assertIsNone(session.get_document("collection1", None))
                self.assertIsNone(session.get_document("collection1", 1))

        def test_remove_document(self):
            """
            Tests the method removing a document
            """
            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding documents
                document = {}
                document["name"] = "document1"
                session.add_document("collection1", document)
                document = {}
                document["name"] = "document2"
                session.add_document("collection1", document)

                # Adding a field
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )
                session.add_field("collection1", "FOV", list[int], None)

                # Adding a value
                document = session["collection1"]["document1"]
                document["PatientName"] = "test"
                document["FOV"] = [1, 2, 3]
                session["collection1"]["document1"] = document

                # Removing a document
                session.remove_document("collection1", "document1")

                # Testing that the document is removed from all tables
                self.assertIsNone(session.get_document("collection1", "document1"))

                # Testing with a collection not existing
                self.assertRaises(
                    ValueError,
                    lambda: session.remove_document(
                        "collection_not_existing", "document1"
                    ),
                )

                # Removing a document
                session.remove_document("collection1", "document2")

                # Testing that the document is removed from the collection
                self.assertIsNone(session.get_document("collection1", "document2"))

        def test_add_document(self):
            """
            Tests the method adding a document
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding fields
                session.add_field("collection1", "List", list[int])
                session.add_field("collection1", "Int", int)

                # Adding a document
                document = {}
                document["name"] = "document1"
                document["List"] = [1, 2, 3]
                document["Int"] = 5
                session.add_document("collection1", document)

                # Testing that the document has been added
                document = session.get_document("collection1", "document1")
                self.assertEqual(document["name"], "document1")

                # Testing when trying to add a document that already exists
                document = {}
                document["name"] = "document1"
                with self.assertRaises(session.database_exceptions):
                    session.add_document("collection1", document)

                # Testing with invalid parameters
                self.assertRaises(
                    ValueError, lambda: session.add_document(15, "document1")
                )
                self.assertRaises(
                    ValueError,
                    lambda: session.add_document(
                        "collection_not_existing", "document1"
                    ),
                )
                self.assertRaises(
                    AttributeError, lambda: session.add_document("collection1", True)
                )

                # Testing the add of several documents
                document = {}
                document["name"] = "document2"
                session.add_document("collection1", document)

                # Adding a document with a dictionary without the primary key
                document = {}
                document["no_primary_key"] = "document1"
                with self.assertRaises(session.database_exceptions):
                    session.add_document("collection1", document)

        def test_add_collection(self):
            """
            Tests the method adding a collection
            """

            database = self.create_database()
            with database as session:
                # Adding a first collection
                session.add_collection("collection1")

                # Checking values
                collection = session.get_collection("collection1")
                self.assertEqual(collection.name, "collection1")
                self.assertEqual(collection.primary_key, {"primary_key": str})

                # Adding a second collection
                session.add_collection("collection2", "id")

                # Checking values
                collection = session.get_collection("collection2")
                self.assertEqual(collection.name, "collection2")
                self.assertEqual(collection.primary_key, {"id": str})

                # Trying with a collection already existing
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.add_collection("collection1"),
                )

                # Trying with table names already taken
                session.add_field("collection1", "test", str, description="Test field")
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.add_collection(populse_db_table),
                )

                # Trying with wrong types
                self.assertRaises(
                    AttributeError,
                    lambda: session.add_collection("collection_valid", True),
                )

        def test_remove_collection(self):
            """
            Tests the method removing a collection
            """

            database = self.create_database()
            with database as session:
                # Adding a first collection
                self.assertFalse(session.has_collection("collection1"))
                session.add_collection("collection1")
                self.assertTrue(session.has_collection("collection1"))

                # Checking values
                collection = session.get_collection("collection1")
                self.assertEqual(collection.name, "collection1")
                self.assertEqual(collection.primary_key, {"primary_key": str})

                # Removing a collection
                session.remove_collection("collection1")
                self.assertFalse(session.has_collection("collection1"))

                # Testing that it does not exist anymore
                self.assertIsNone(session.get_collection("collection1"))

                # Adding new collections
                session.add_collection("collection1")
                session.add_collection("collection2")

                # Checking values
                collection = session.get_collection("collection1")
                self.assertEqual(collection.name, "collection1")
                self.assertEqual(collection.primary_key, {"primary_key": str})
                collection = session.get_collection("collection2")
                self.assertEqual(collection.name, "collection2")
                self.assertEqual(collection.primary_key, {"primary_key": str})

                # Removing one collection and testing that the other is unchanged
                session.remove_collection("collection2")
                collection = session.get_collection("collection1")
                self.assertEqual(collection.name, "collection1")
                self.assertEqual(collection.primary_key, {"primary_key": str})
                self.assertIsNone(session.get_collection("collection2"))

                # Adding a field
                session.add_field("collection1", "Field", str)
                field = session.get_field("collection1", "Field")
                self.assertEqual(field["name"], "Field")
                self.assertEqual(field["collection"], "collection1")
                self.assertIsNone(field["description"])
                self.assertEqual(field["type"], str)

                # Adding a document
                session["collection1"]["document"] = {}
                document = session.get_document("collection1", "document")
                self.assertEqual(document["primary_key"], "document")

                # Removing the collection containing the field and the document and testing that it is indeed removed
                session.remove_collection("collection1")
                self.assertIsNone(session.get_collection("collection1"))
                self.assertIsNone(session.get_field("collection1", "name"))
                self.assertIsNone(session.get_field("collection1", "Field"))
                self.assertIsNone(session.get_document("collection1", "document"))

                # Testing with a collection not existing
                self.assertRaises(
                    session.database_exceptions,
                    lambda: session.remove_collection("collection_not_existing"),
                )
                self.assertRaises(
                    session.database_exceptions, lambda: session.remove_collection(True)
                )

        def test_get_collection(self):
            """
            Tests the method giving the collection row
            """

            database = self.create_database()
            with database as session:
                # Adding a first collection
                session.add_collection("collection1")

                # Checking values
                collection = session.get_collection("collection1")
                self.assertEqual(collection.name, "collection1")
                self.assertEqual(collection.primary_key, {"primary_key": str})

                # Adding a second collection
                session.add_collection("collection2", "id")

                # Checking values
                collection = session.get_collection("collection2")
                self.assertEqual(collection.name, "collection2")
                self.assertEqual(collection.primary_key, {"id": str})

                # Trying with a collection not existing
                self.assertIsNone(session.get_collection("collection3"))

        def test_get_collections(self):
            """
            Tests the method giving the list of collections rows
            """

            database = self.create_database()
            with database as session:
                # Testing that there is no collection at first
                self.assertEqual(list(session.get_collections()), [])

                # Adding a collection
                session.add_collection("collection1")

                collections = list(session.get_collections())
                self.assertEqual(len(collections), 1)
                self.assertEqual(collections[0].name, "collection1")

                session.add_collection("collection2")

                collections = list(session.get_collections())
                self.assertEqual(len(collections), 2)

                session.remove_collection("collection2")

                collections = list(session.get_collections())
                self.assertEqual(len(collections), 1)
                self.assertEqual(collections[0].name, "collection1")

        def test_get_collections_names(self):
            """
            Tests the method giving the collections names
            """

            database = self.create_database()
            with database as session:
                # Testing that there is no collection at first
                self.assertEqual(list(session.get_collections_names()), [])

                # Adding a collection
                session.add_collection("collection1")

                self.assertEqual(list(session.get_collections_names()), ["collection1"])

                session.add_collection("collection2")

                collections = list(session.get_collections_names())
                self.assertEqual(len(collections), 2)
                self.assertTrue("collection1" in collections)
                self.assertTrue("collection2" in collections)

                session.remove_collection("collection2")

                self.assertEqual(list(session.get_collections_names()), ["collection1"])

        def test_get_documents(self):
            """
            Tests the method returning the list of document rows, given a collection
            """

            database = self.create_database()
            with database as session:
                # Adding collections
                session.add_collection("collection1", "name")
                session.add_collection("collection2", "id")

                session["collection1"]["document1"] = {}
                session["collection1"]["document2"] = {}
                session["collection2"]["document1"] = {}
                session["collection2"]["document2"] = {}

                documents1 = list(session.get_documents("collection1"))
                self.assertEqual(len(documents1), 2)

                documents2 = list(session.get_documents("collection2"))
                self.assertEqual(len(documents2), 2)

                # Testing with a collection not existing
                self.assertEqual(
                    list(session.get_documents("collection_not_existing")), []
                )
                self.assertEqual(list(session.get_documents("collection")), [])
                self.assertEqual(list(session.get_documents(None)), [])

        def test_get_documents_names(self):
            """
            Tests the method returning the list of document names, given a collection
            """

            database = self.create_database()
            with database as session:
                # Adding collections
                session.add_collection("collection1", "name")
                session.add_collection("collection2", "FileName")

                session["collection1"]["document1"] = {}
                session["collection1"]["document2"] = {}
                session["collection2"]["document3"] = {}
                session["collection2"]["document4"] = {}

                documents1 = list(session.get_documents_ids("collection1"))
                self.assertEqual(len(documents1), 2)
                self.assertTrue(["document1"] in documents1)
                self.assertTrue(["document2"] in documents1)

                documents2 = list(session.get_documents_ids("collection2"))
                self.assertEqual(len(documents2), 2)
                self.assertTrue(["document3"] in documents2)
                self.assertTrue(["document4"] in documents2)

                # Testing with a collection not existing
                self.assertEqual(
                    list(session.get_documents_ids("collection_not_existing")), []
                )
                self.assertEqual(list(session.get_documents_ids("collection")), [])
                self.assertEqual(list(session.get_documents_ids(None)), [])

        def test_remove_value(self):
            """
            Tests the method removing a value
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding fields
                session.add_field(
                    "collection1", "PatientName", str, "Name of the patient"
                )
                session.add_field("collection1", "Bits per voxel", int)
                session.add_field("collection1", "Dataset dimensions", list[int], None)

                # Adding a document
                doc = {
                    "PatientName": "test",
                    "Dataset dimensions": [3, 28, 28, 3],
                    "Bits per voxel": 42,
                }
                session["collection1"]["document1"] = doc

                doc["Bits per voxel"] = "space_field"
                # SQLite allows to store any type in columns
                # with self.assertRaises(TypeError):
                session["collection1"]["document1"] = doc
                value = session["collection1"]["document1"]["Dataset dimensions"]
                self.assertEqual(value, [3, 28, 28, 3])

                # Removing values
                del doc["PatientName"]
                del doc["Bits per voxel"]
                del doc["Dataset dimensions"]
                session["collection1"]["document1"] = doc

                # Testing that the values are actually removed
                doc = session["collection1"]["document1"]
                self.assertIsNone(doc.get("PatientName"))
                self.assertIsNone(doc.get("Bits per voxel"))
                self.assertIsNone(doc.get("Dataset dimensions"))

        def test_list_dates(self):
            """
            Tests the storage and retrieval of fields of type list of time, date
            and datetime
            """

            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding fields
                session.add_field("collection1", "list_date", list[date], None)
                session.add_field("collection1", "list_time", list[time], None)
                session.add_field("collection1", "list_datetime", list[datetime], None)

                document = {}
                document["name"] = "document1"
                session.add_document("collection1", document)

                list_date = [date(2018, 5, 23), date(1899, 12, 31)]
                list_time = [time(12, 41, 33, 540), time(1, 2, 3)]
                list_datetime = [
                    datetime(2018, 5, 23, 12, 41, 33, 540),
                    datetime(1899, 12, 31, 1, 2, 3),
                ]

                document["list_date"] = list_date
                document["list_time"] = list_time
                document["list_datetime"] = list_datetime
                session["collection1"]["document1"] = document

                document = session["collection1"]["document1"]
                self.assertEqual(list_date, document["list_date"])
                self.assertEqual(list_time, document["list_time"])
                self.assertEqual(list_datetime, document["list_datetime"])

        def test_json_field(self):
            """
            Tests the storage and retrieval of fields of type JSON
            """

            doc = {"name": "the_name", "json": {"key": [1, 2, "three"]}}
            database = self.create_database()
            with database as session:
                # Adding a collection
                session.add_collection("collection1", "name")

                # Adding fields
                session.add_field("collection1", "json", dict)

                session.add_document("collection1", doc)
                self.assertEqual(doc, session.get_document("collection1", "the_name"))
                self.assertIsNone(
                    session.get_document("collection1", "not_a_valid_name")
                )
            with database as session:
                self.assertEqual(doc, session.get_document("collection1", "the_name"))
                self.assertIsNone(
                    session.get_document("collection1", "not_a_valid_name")
                )

        def test_filter_documents(self):
            """
            Tests the method applying the filter
            """

            database = self.create_database()
            with database as session:
                session.add_collection("collection_test")
                session.add_field("collection_test", "field_test", str, None)
                session["collection_test"]["document_test"] = {}

                # Checking with invalid collection
                self.assertRaises(
                    ValueError,
                    lambda: {
                        document["primary_key"]
                        for document in session.filter_documents(
                            "collection_not_existing", None
                        )
                    },
                )

                # Checking that every document is returned if there is no filter
                documents = {
                    document["primary_key"]
                    for document in session.filter_documents("collection_test", None)
                }
                self.assertEqual(documents, {"document_test"})

        def test_filters(self):
            list_datetime = [
                datetime(2018, 5, 23, 12, 41, 33, 540),
                datetime(1981, 5, 8, 20, 0),
                datetime(1899, 12, 31, 1, 2, 3),
            ]

            database = self.create_database()
            with database as session:
                session.add_collection("collection1", "name")

                session.add_field(
                    "collection1",
                    "format",
                    field_type=str,
                    description=None,
                    index=True,
                )
                session.add_field(
                    "collection1", "strings", field_type=list[str], description=None
                )
                session.add_field(
                    "collection1", "datetime", field_type=datetime, description=None
                )
                session.add_field(
                    "collection1", "has_format", field_type=bool, description=None
                )

                files = ("abc", "bcd", "def", "xyz")
                for file in files:
                    for dt in list_datetime:
                        for format, ext in (
                            ("NIFTI", "nii"),
                            ("DICOM", "dcm"),
                            ("Freesurfer", "mgz"),
                        ):
                            document = dict(
                                name=f"/{file}_{dt.year}.{ext}",
                                format=format,
                                strings=list(file),
                                datetime=dt,
                                has_format=True,
                            )
                            session.add_document("collection1", document)
                        document = f"/{file}_{dt.year}.none"
                        d = dict(name=document, strings=list(file))
                        session.add_document("collection1", d)

                assert (
                    list(session.filter_documents("collection1", "format IN []")) == []
                )

                for filter, expected in (
                    (
                        'format == "NIFTI"',
                        {
                            "/xyz_1899.nii",
                            "/xyz_2018.nii",
                            "/abc_2018.nii",
                            "/bcd_1899.nii",
                            "/bcd_2018.nii",
                            "/def_1899.nii",
                            "/abc_1981.nii",
                            "/def_2018.nii",
                            "/def_1981.nii",
                            "/bcd_1981.nii",
                            "/abc_1899.nii",
                            "/xyz_1981.nii",
                        },
                    ),
                    (
                        '"b" IN strings',
                        {
                            "/bcd_2018.mgz",
                            "/abc_1899.mgz",
                            "/abc_1899.dcm",
                            "/bcd_1981.dcm",
                            "/abc_1981.dcm",
                            "/bcd_1981.mgz",
                            "/bcd_1899.mgz",
                            "/abc_1981.mgz",
                            "/abc_2018.mgz",
                            "/abc_2018.dcm",
                            "/bcd_2018.dcm",
                            "/bcd_1899.dcm",
                            "/abc_2018.nii",
                            "/bcd_1899.nii",
                            "/abc_1981.nii",
                            "/bcd_1981.nii",
                            "/abc_1899.nii",
                            "/bcd_2018.nii",
                            "/abc_1899.none",
                            "/bcd_1899.none",
                            "/bcd_1981.none",
                            "/abc_2018.none",
                            "/bcd_2018.none",
                            "/abc_1981.none",
                        },
                    ),
                    (
                        '(format == "NIFTI" OR NOT format == "DICOM")',
                        {
                            "/xyz_1899.nii",
                            "/xyz_1899.mgz",
                            "/bcd_2018.mgz",
                            "/bcd_1899.nii",
                            "/bcd_2018.nii",
                            "/def_1899.nii",
                            "/bcd_1981.mgz",
                            "/abc_1981.nii",
                            "/def_2018.mgz",
                            "/abc_1899.nii",
                            "/def_1899.mgz",
                            "/xyz_1899.none",
                            "/abc_2018.nii",
                            "/def_1899.none",
                            "/bcd_1899.mgz",
                            "/def_2018.nii",
                            "/abc_1981.mgz",
                            "/abc_1899.none",
                            "/xyz_1981.mgz",
                            "/bcd_1981.nii",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/def_1981.nii",
                            "/def_1981.mgz",
                            "/bcd_1899.none",
                            "/xyz_2018.mgz",
                            "/bcd_1981.none",
                            "/xyz_1981.none",
                            "/abc_1981.none",
                            "/def_2018.none",
                            "/xyz_2018.none",
                            "/abc_2018.none",
                            "/def_1981.none",
                            "/bcd_2018.none",
                        },
                    ),
                    (
                        '"a" IN strings',
                        {
                            "/abc_1899.none",
                            "/abc_1899.nii",
                            "/abc_2018.nii",
                            "/abc_1899.mgz",
                            "/abc_1899.dcm",
                            "/abc_1981.dcm",
                            "/abc_1981.nii",
                            "/abc_1981.mgz",
                            "/abc_2018.mgz",
                            "/abc_2018.dcm",
                            "/abc_2018.none",
                            "/abc_1981.none",
                        },
                    ),
                    (
                        'NOT "b" IN strings',
                        {
                            "/xyz_1899.nii",
                            "/xyz_2018.dcm",
                            "/def_1981.dcm",
                            "/xyz_2018.nii",
                            "/xyz_1981.dcm",
                            "/def_1899.none",
                            "/xyz_1899.dcm",
                            "/xyz_1981.nii",
                            "/def_1899.dcm",
                            "/def_1899.nii",
                            "/def_2018.mgz",
                            "/def_2018.nii",
                            "/xyz_1899.mgz",
                            "/def_2018.dcm",
                            "/def_1899.mgz",
                            "/def_1981.mgz",
                            "/xyz_1981.mgz",
                            "/xyz_2018.mgz",
                            "/xyz_1899.none",
                            "/def_1981.nii",
                            "/xyz_2018.none",
                            "/xyz_1981.none",
                            "/def_2018.none",
                            "/def_1981.none",
                        },
                    ),
                    (
                        '("a" IN strings OR NOT "b" IN strings)',
                        {
                            "/xyz_1899.nii",
                            "/xyz_1899.mgz",
                            "/def_1899.nii",
                            "/abc_1981.nii",
                            "/def_2018.mgz",
                            "/abc_1899.nii",
                            "/def_1899.mgz",
                            "/abc_2018.dcm",
                            "/xyz_1899.none",
                            "/xyz_2018.dcm",
                            "/def_1981.dcm",
                            "/abc_2018.nii",
                            "/def_1899.none",
                            "/abc_1981.dcm",
                            "/def_2018.nii",
                            "/abc_1981.mgz",
                            "/def_2018.dcm",
                            "/abc_1899.none",
                            "/xyz_1981.mgz",
                            "/xyz_1899.dcm",
                            "/abc_1899.dcm",
                            "/def_1899.dcm",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/xyz_1981.dcm",
                            "/def_1981.nii",
                            "/def_1981.mgz",
                            "/xyz_2018.mgz",
                            "/xyz_1981.none",
                            "/abc_1981.none",
                            "/def_2018.none",
                            "/xyz_2018.none",
                            "/abc_2018.none",
                            "/def_1981.none",
                        },
                    ),
                    (
                        'format IN ["DICOM", "NIFTI"]',
                        {
                            "/xyz_1899.nii",
                            "/xyz_2018.dcm",
                            "/bcd_1899.nii",
                            "/def_1899.nii",
                            "/abc_1981.nii",
                            "/abc_1899.nii",
                            "/bcd_2018.nii",
                            "/abc_2018.dcm",
                            "/bcd_1899.dcm",
                            "/def_1981.dcm",
                            "/abc_2018.nii",
                            "/abc_1981.dcm",
                            "/bcd_2018.dcm",
                            "/def_2018.nii",
                            "/def_2018.dcm",
                            "/xyz_1899.dcm",
                            "/abc_1899.dcm",
                            "/def_1899.dcm",
                            "/bcd_1981.nii",
                            "/xyz_1981.nii",
                            "/xyz_2018.nii",
                            "/xyz_1981.dcm",
                            "/def_1981.nii",
                            "/bcd_1981.dcm",
                        },
                    ),
                    (
                        '(format == "NIFTI" OR NOT format == "DICOM") AND ("a" IN strings OR NOT "b" IN strings)',
                        {
                            "/abc_1899.none",
                            "/xyz_1899.mgz",
                            "/abc_1981.mgz",
                            "/abc_2018.nii",
                            "/xyz_1899.nii",
                            "/abc_1899.mgz",
                            "/def_1899.mgz",
                            "/def_1899.nii",
                            "/def_1899.none",
                            "/abc_1981.nii",
                            "/def_2018.nii",
                            "/xyz_2018.nii",
                            "/def_1981.nii",
                            "/abc_1899.nii",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/def_1981.mgz",
                            "/xyz_2018.mgz",
                            "/xyz_1899.none",
                            "/def_2018.mgz",
                            "/xyz_1981.mgz",
                            "/xyz_1981.none",
                            "/abc_1981.none",
                            "/def_2018.none",
                            "/xyz_2018.none",
                            "/abc_2018.none",
                            "/def_1981.none",
                        },
                    ),
                    (
                        'format > "DICOM"',
                        {
                            "/xyz_1899.nii",
                            "/xyz_1899.mgz",
                            "/bcd_2018.mgz",
                            "/bcd_1899.nii",
                            "/bcd_2018.nii",
                            "/def_1899.nii",
                            "/bcd_1981.mgz",
                            "/abc_1981.nii",
                            "/def_2018.mgz",
                            "/abc_1899.nii",
                            "/def_1899.mgz",
                            "/abc_2018.nii",
                            "/def_2018.nii",
                            "/abc_1981.mgz",
                            "/xyz_1981.mgz",
                            "/bcd_1981.nii",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/def_1981.nii",
                            "/def_1981.mgz",
                            "/bcd_1899.mgz",
                            "/xyz_2018.mgz",
                        },
                    ),
                    (
                        'format <= "DICOM"',
                        {
                            "/abc_1981.dcm",
                            "/def_1899.dcm",
                            "/abc_2018.dcm",
                            "/bcd_1899.dcm",
                            "/def_1981.dcm",
                            "/bcd_2018.dcm",
                            "/def_2018.dcm",
                            "/xyz_2018.dcm",
                            "/xyz_1899.dcm",
                            "/abc_1899.dcm",
                            "/xyz_1981.dcm",
                            "/bcd_1981.dcm",
                        },
                    ),
                    (
                        'format > "DICOM" AND strings != ["b", "c", "d"]',
                        {
                            "/xyz_1899.nii",
                            "/xyz_1899.mgz",
                            "/abc_1981.mgz",
                            "/abc_2018.nii",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/def_1899.mgz",
                            "/def_1899.nii",
                            "/abc_1981.nii",
                            "/def_2018.nii",
                            "/def_1981.nii",
                            "/abc_1899.nii",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/def_1981.mgz",
                            "/xyz_2018.mgz",
                            "/def_2018.mgz",
                            "/xyz_1981.mgz",
                        },
                    ),
                    (
                        'format <= "DICOM" AND strings == ["b", "c", "d"]',
                        {
                            "/bcd_2018.dcm",
                            "/bcd_1981.dcm",
                            "/bcd_1899.dcm",
                        },
                    ),
                    (
                        "has_format in [false, null]",
                        {
                            "/def_1899.none",
                            "/abc_1899.none",
                            "/bcd_1899.none",
                            "/xyz_1899.none",
                            "/bcd_2018.none",
                            "/abc_1981.none",
                            "/def_2018.none",
                            "/xyz_2018.none",
                            "/abc_2018.none",
                            "/def_1981.none",
                            "/xyz_1981.none",
                            "/bcd_1981.none",
                        },
                    ),
                    (
                        "format == null",
                        {
                            "/bcd_1981.none",
                            "/abc_1899.none",
                            "/def_1899.none",
                            "/bcd_2018.none",
                            "/abc_1981.none",
                            "/def_2018.none",
                            "/xyz_2018.none",
                            "/abc_2018.none",
                            "/def_1981.none",
                            "/bcd_1899.none",
                            "/xyz_1899.none",
                            "/xyz_1981.none",
                        },
                    ),
                    ("strings == null", set()),
                    (
                        "strings != NULL",
                        {
                            "/xyz_1899.nii",
                            "/xyz_2018.dcm",
                            "/xyz_1899.mgz",
                            "/bcd_2018.mgz",
                            "/bcd_1899.nii",
                            "/def_2018.none",
                            "/def_1899.mgz",
                            "/def_1899.nii",
                            "/bcd_1981.mgz",
                            "/abc_1981.nii",
                            "/def_2018.mgz",
                            "/abc_1899.nii",
                            "/bcd_2018.nii",
                            "/abc_2018.dcm",
                            "/xyz_1899.none",
                            "/bcd_1899.dcm",
                            "/bcd_1981.none",
                            "/def_1981.dcm",
                            "/abc_2018.nii",
                            "/def_1899.none",
                            "/xyz_1981.none",
                            "/abc_1981.dcm",
                            "/bcd_2018.dcm",
                            "/def_2018.nii",
                            "/abc_1981.mgz",
                            "/def_2018.dcm",
                            "/abc_1899.none",
                            "/xyz_1981.mgz",
                            "/xyz_1899.dcm",
                            "/abc_1899.dcm",
                            "/def_1899.dcm",
                            "/bcd_1981.nii",
                            "/def_1981.none",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/xyz_2018.none",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/bcd_1899.mgz",
                            "/bcd_2018.none",
                            "/abc_1981.none",
                            "/xyz_1981.dcm",
                            "/abc_2018.none",
                            "/def_1981.nii",
                            "/bcd_1981.dcm",
                            "/def_1981.mgz",
                            "/bcd_1899.none",
                            "/xyz_2018.mgz",
                        },
                    ),
                    (
                        "format != NULL",
                        {
                            "/xyz_1899.nii",
                            "/xyz_1899.mgz",
                            "/bcd_2018.mgz",
                            "/bcd_1899.nii",
                            "/def_1899.mgz",
                            "/def_1899.nii",
                            "/bcd_1981.mgz",
                            "/abc_1981.nii",
                            "/def_2018.mgz",
                            "/abc_1899.nii",
                            "/bcd_2018.nii",
                            "/abc_2018.dcm",
                            "/xyz_1981.mgz",
                            "/def_1981.dcm",
                            "/abc_2018.nii",
                            "/abc_1981.dcm",
                            "/bcd_2018.dcm",
                            "/def_2018.nii",
                            "/bcd_1981.nii",
                            "/abc_1981.mgz",
                            "/def_2018.dcm",
                            "/bcd_1899.dcm",
                            "/xyz_2018.dcm",
                            "/xyz_1899.dcm",
                            "/abc_1899.dcm",
                            "/def_1899.dcm",
                            "/bcd_1899.mgz",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/xyz_1981.dcm",
                            "/def_1981.nii",
                            "/bcd_1981.dcm",
                            "/def_1981.mgz",
                            "/xyz_2018.mgz",
                        },
                    ),
                    (
                        'name like "%.nii"',
                        {
                            "/xyz_1899.nii",
                            "/xyz_2018.nii",
                            "/abc_2018.nii",
                            "/bcd_1899.nii",
                            "/bcd_2018.nii",
                            "/def_1899.nii",
                            "/abc_1981.nii",
                            "/def_2018.nii",
                            "/def_1981.nii",
                            "/bcd_1981.nii",
                            "/abc_1899.nii",
                            "/xyz_1981.nii",
                        },
                    ),
                    (
                        'name ilike "%A%"',
                        {
                            "/abc_1899.none",
                            "/abc_1899.nii",
                            "/abc_2018.nii",
                            "/abc_1899.mgz",
                            "/abc_1899.dcm",
                            "/abc_1981.dcm",
                            "/abc_1981.nii",
                            "/abc_1981.mgz",
                            "/abc_2018.mgz",
                            "/abc_2018.dcm",
                            "/abc_2018.none",
                            "/abc_1981.none",
                        },
                    ),
                    (
                        "all",
                        {
                            "/xyz_1899.nii",
                            "/xyz_2018.dcm",
                            "/xyz_1899.mgz",
                            "/bcd_2018.mgz",
                            "/bcd_1899.nii",
                            "/def_2018.none",
                            "/def_1899.mgz",
                            "/def_1899.nii",
                            "/bcd_1981.mgz",
                            "/abc_1981.nii",
                            "/def_2018.mgz",
                            "/abc_1899.nii",
                            "/bcd_2018.nii",
                            "/abc_2018.dcm",
                            "/xyz_1899.none",
                            "/bcd_1899.dcm",
                            "/bcd_1981.none",
                            "/def_1981.dcm",
                            "/abc_2018.nii",
                            "/def_1899.none",
                            "/xyz_1981.none",
                            "/abc_1981.dcm",
                            "/bcd_2018.dcm",
                            "/def_2018.nii",
                            "/abc_1981.mgz",
                            "/def_2018.dcm",
                            "/abc_1899.none",
                            "/xyz_1981.mgz",
                            "/xyz_1899.dcm",
                            "/abc_1899.dcm",
                            "/def_1899.dcm",
                            "/bcd_1981.nii",
                            "/def_1981.none",
                            "/xyz_1981.nii",
                            "/abc_2018.mgz",
                            "/xyz_2018.none",
                            "/xyz_2018.nii",
                            "/abc_1899.mgz",
                            "/bcd_1899.mgz",
                            "/bcd_2018.none",
                            "/abc_1981.none",
                            "/xyz_1981.dcm",
                            "/abc_2018.none",
                            "/def_1981.nii",
                            "/bcd_1981.dcm",
                            "/def_1981.mgz",
                            "/bcd_1899.none",
                            "/xyz_2018.mgz",
                        },
                    ),
                ):
                    for tested_filter in (
                        filter,
                        f"({filter}) AND ALL",
                        f"ALL AND ({filter})",
                    ):
                        try:
                            documents = {
                                document["name"]
                                for document in session.filter_documents(
                                    "collection1", tested_filter
                                )
                            }
                            self.assertEqual(documents, expected)
                        except Exception as e:
                            raise Exception(
                                f"Error while testing filter : {tested_filter}"
                            ) from e
                    all_documents = {
                        document["name"]
                        for document in session.filter_documents("collection1", "ALL")
                    }
                    for tested_filter in (
                        f"({filter}) OR ALL",
                        f"ALL OR ({filter})",
                    ):
                        try:
                            documents = {
                                document["name"]
                                for document in session.filter_documents(
                                    "collection1", tested_filter
                                )
                            }
                            self.assertEqual(documents, all_documents)
                        except Exception as e:
                            raise Exception(
                                f"Error while testing filter : {tested_filter}"
                            ) from e

        def test_modify_list_field(self):
            database = self.create_database()
            with database as session:
                session.add_collection("collection1", "name")
                session.add_field(
                    "collection1", "strings", field_type=list[str], description=None
                )
                session["collection1"]["test"] = {"strings": ["a", "b", "c"]}
                names = list(
                    document["name"]
                    for document in session.filter_documents(
                        "collection1", '"b" IN strings'
                    )
                )
                self.assertEqual(names, ["test"])

                session["collection1"]["test"] = {"strings": ["x", "y", "z"]}
                names = list(
                    document["name"]
                    for document in session.filter_documents(
                        "collection1", '"b" IN strings'
                    )
                )
                self.assertEqual(names, [])
                names = list(
                    document["name"]
                    for document in session.filter_documents(
                        "collection1", '"z" IN strings'
                    )
                )
                self.assertEqual(names, ["test"])

                session["collection1"]["test"] = {}
                names = list(
                    document["name"]
                    for document in session.filter_documents(
                        "collection1", '"y" IN strings'
                    )
                )
                self.assertEqual(names, [])

        def test_filter_literals(self):
            """
            Test the Python values returned (internally) for literals by the
            interpreter of filter expression
            """

            literals = {
                "True": True,
                "TRUE": True,
                "true": True,
                "False": False,
                "FALSE": False,
                "false": False,
                "Null": None,
                "NULL": None,
                "null": None,
                "0": 0,
                "123456789101112": 123456789101112,
                "-45": -45,
                "-46.8": -46.8,
                "1.5654353456363e-15": 1.5654353456363e-15,
                '""': "",
                '"2018-05-25"': "2018-05-25",
                '"a\n b\n  c"': "a\n b\n  c",
                '"\\""': '"',
                "2018-05-25": date(2018, 5, 25),
                "2018-5-25": date(2018, 5, 25),
                "12:54": time(12, 54),
                "02:4:9": time(2, 4, 9),
                # The following interpretation of microsecond is a strange
                # behavior of datetime.strptime that expect up to 6 digits
                # with zeroes padded on the right !?
                "12:34:56.789": time(12, 34, 56, 789000),
                "12:34:56.000789": time(12, 34, 56, 789),
                "2018-05-25T12:34:56.000789": datetime(2018, 5, 25, 12, 34, 56, 789),
                "2018-5-25T12:34": datetime(2018, 5, 25, 12, 34),
                "[]": [],
            }
            # Adds the literal for a list of all elements in the dictionary
            literals[f"[{','.join(literals.keys())}]"] = list(literals.values())

            parser = literal_parser()
            for literal, expected_value in literals.items():
                tree = parser.parse(literal)
                value = FilterToSQL(None).transform(tree)
                self.assertEqual(value, expected_value)

        def test_with(self):
            """
            Tests the database session
            """

            database = self.create_database()
            try:
                with database as session:
                    session.add_collection("collection1", "name")
                    session.add_document("collection1", {"name": "toto"})
                    # Raise an exception to roll back modifications
                    boom  # noqa: B018
            except NameError:
                pass

            with database as session:
                session.add_collection("collection1", "name")
                session.add_document("collection1", {"name": "titi"})

            # Reopen the database to check that "titi" was committed
            database = self.create_database(clear=False)
            with database as session:
                self.assertTrue(session.has_collection("collection1"))
                names = [
                    i["name"] for i in session.filter_documents("collection1", "all")
                ]
                self.assertEqual(names, ["titi"])

                # Check that recursive session creation always return the
                # same object
                with database as session2:
                    self.assertIs(session, session2)
                    with database as session3:
                        self.assertIs(session, session3)
                        session.add_document("collection1", {"name": "toto"})

            # Check that previous session was committed and released.
            with database as session4:
                self.assertIsNot(session, session4)

            # Destroy the database and create a new one
            database = self.create_database(clear=False)

            # Check that previous session was destroyed and that
            # a new one is created.
            with database as session5:
                self.assertIsNot(session, session5)
                self.assertEqual(len(list(session5.get_documents("collection1"))), 2)

        def test_automatic_fields_creation(self):
            """
            Test automatic creation of fields with add_document
            """
            database = self.create_database()
            with database as session:
                now = datetime.now()
                session.add_collection("test")
                base_doc = {
                    "string": "string",
                    "int": 1,
                    "float": 1.4,
                    "boolean": True,
                    "datetime": now,
                    "date": now.date(),
                    "time": now.time(),
                    "dict": {
                        "string": "string",
                        "int": 1,
                        "float": 1.4,
                        "boolean": True,
                    },
                }
                doc = base_doc.copy()
                for k, v in base_doc.items():
                    lk = f"list_{k}"
                    doc[lk] = [v]
                doc["primary_key"] = "test"
                session.add_document("test", doc)
                self.maxDiff = None
                stored_doc = session.get_document("test", "test")
                self.assertEqual(doc, stored_doc)

        def test_delete(self):
            """
            Test automatic creation of fields with add_document
            """
            database = self.create_database()
            with database as session:
                session.add_collection("things", ("one", "two"))
                for one in range(10):
                    for two in range(10):
                        session["things"][(str(one), str(two))] = {
                            "content": one * two,
                        }
                deleted = session["things"].delete("all")
                self.assertEqual(deleted, 100)
                deleted = session["things"].delete("{content} >= 4")
                self.assertEqual(deleted, 0)
                for one in range(10):
                    for two in range(10):
                        session["things"][(str(one), str(two))] = {
                            "content": one * two,
                        }
                deleted = session["things"].delete('{one} == "5"')
                self.assertEqual(deleted, 10)
                deleted = session["things"].delete("{content} >= 4")
                self.assertEqual(deleted, 67)
                self.maxDiff = 1000
                self.assertEqual(
                    list(session["things"].documents()),
                    [
                        {"content": 0, "one": 0, "two": 0},
                        {"content": 0, "one": 0, "two": 1},
                        {"content": 0, "one": 0, "two": 2},
                        {"content": 0, "one": 0, "two": 3},
                        {"content": 0, "one": 0, "two": 4},
                        {"content": 0, "one": 0, "two": 5},
                        {"content": 0, "one": 0, "two": 6},
                        {"content": 0, "one": 0, "two": 7},
                        {"content": 0, "one": 0, "two": 8},
                        {"content": 0, "one": 0, "two": 9},
                        {"content": 0, "one": 1, "two": 0},
                        {"content": 1, "one": 1, "two": 1},
                        {"content": 2, "one": 1, "two": 2},
                        {"content": 3, "one": 1, "two": 3},
                        {"content": 0, "one": 2, "two": 0},
                        {"content": 2, "one": 2, "two": 1},
                        {"content": 0, "one": 3, "two": 0},
                        {"content": 3, "one": 3, "two": 1},
                        {"content": 0, "one": 4, "two": 0},
                        {"content": 0, "one": 6, "two": 0},
                        {"content": 0, "one": 7, "two": 0},
                        {"content": 0, "one": 8, "two": 0},
                        {"content": 0, "one": 9, "two": 0},
                    ],
                )

    return TestDatabaseMethods


TestDatabaseMethods = create_test_case()

# def load_tests(loader, standard_tests, pattern):
#     """
#     Prepares the tests parameters

#     :param loader:

#     :param standard_tests:

#     :param pattern:

#     :return: A test suite
#     """
#     suite = unittest.TestSuite()
#     suite.addTests(loader.loadTestsFromTestCase(TestsSQLiteInMemory))
#     tests = loader.loadTestsFromTestCase(create_test_case())
#     suite.addTests(tests)

#     # Tests with postgresql. All the tests will be skipped if
#     # it is not possible to connect to populse_db_tests database.
#     # tests = loader.loadTestsFromTestCase(create_test_case(
#     # database_url='postgresql:///populse_db_tests',
#     # caches=False,
#     # list_tables=True,
#     # query_type='mixed'))
#     # suite.addTests(tests)

#     return suite



schemas = [
    {
        "version": "1.0.1",
        "schema": {
            "snapshots": [
                {
                    "subject": ["str", {"primary_key": True}],
                    "time_point": [str, {"primary_key": True}],
                    "data_type": ["str", {"primary_key": True}],
                    "image": str,
                    "top": "list[int]",
                    "size": list[float],
                    "side": "str",
                    "execution": [str, {"index": True}],
                }
            ],
        },
    },
    {
        "version": "1.0.0",
        "schema": {
            "snapshots": [
                {
                    "subject": ["str", {"primary_key": True}],
                    "time_point": [str, {"primary_key": True}],
                    "data_type": ["str", {"primary_key": True}],
                    "image": str,
                    "top": "list[int]",
                    "size": list[float],
                    "side": "str",
                }
            ],
        },
    },
]
